# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the p2p stub connection."""

import os
import tempfile
from pathlib import Path
from typing import Union, cast

from aea.configurations.base import ConnectionConfig, PublicId
from aea.identity.base import Identity
from aea.mail.base import Envelope

from packages.fetchai.connections.stub.connection import StubConnection, write_envelope


PUBLIC_ID = PublicId.from_str("fetchai/p2p_stub:0.9.0")


class P2PStubConnection(StubConnection):
    r"""A p2p stub connection.

    This connection uses an existing directory as a Rendez-Vous point for agents to communicate locally.
    Each connected agent will create a file named after its address/identity where it can receive messages.

    The connection detects new messages by watchdogging the input file looking for new lines.
    """

    connection_id = PUBLIC_ID

    def __init__(self, configuration: ConnectionConfig, identity: Identity, **kwargs):
        """
        Initialize a p2p stub connection.

        :param configuration: the connection configuration
        :param identity: the identity
        """
        namespace_dir_path = cast(
            Union[str, Path],
            configuration.config.get("namespace_dir", tempfile.mkdtemp()),
        )
        if namespace_dir_path is None:
            raise ValueError("namespace_dir_path must be set!")  # pragma: nocover
        self.namespace = os.path.abspath(namespace_dir_path)

        input_file_path = os.path.join(self.namespace, "{}.in".format(identity.address))
        output_file_path = os.path.join(
            self.namespace, "{}.out".format(identity.address)
        )
        configuration.config["input_file"] = input_file_path
        configuration.config["output_file"] = output_file_path
        super().__init__(configuration=configuration, identity=identity, **kwargs)

    async def send(self, envelope: Envelope):
        """
        Send messages.

        :return: None
        """
        if self.loop is None:
            raise ValueError("Loop not initialized.")  # pragma: nocover
        self._ensure_valid_envelope_for_external_comms(envelope)
        target_file = Path(os.path.join(self.namespace, "{}.in".format(envelope.to)))

        with open(target_file, "ab") as file:
            await self.loop.run_in_executor(
                self._write_pool, write_envelope, envelope, file
            )

    async def disconnect(self) -> None:
        """Disconnect the connection."""
        if self.loop is None:
            raise ValueError("Loop not initialized.")  # pragma: nocover
        await self.loop.run_in_executor(self._write_pool, self._cleanup)
        await super().disconnect()

    def _cleanup(self):
        try:
            os.unlink(self.configuration.config["input_file"])
        except OSError:
            pass
        try:
            os.unlink(self.configuration.config["output_file"])
        except OSError:
            pass
        try:
            os.rmdir(self.namespace)
        except OSError:
            pass



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2020 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests for the sym link module."""

import logging
from unittest.mock import Mock, patch

from aea.helpers.temp_error_handler import ErrorHandler


_default_logger = logging.getLogger(__name__)


def test_send_unsupported_protocol():
    """Test the send_unsupported_protocol method."""
    handler = ErrorHandler
    envelope_mock = Mock()
    envelope_mock.protocol_id = "1"
    envelope_mock.sender = "2"
    envelope_mock.to = "3"
    count = handler.unsupported_protocol_count
    with patch.object(_default_logger, "warning") as mock_logger:
        handler.send_unsupported_protocol(envelope_mock, _default_logger)
        mock_logger.assert_any_call(
            f"Unsupported protocol: {envelope_mock.protocol_id}. You might want to add a handler for this protocol. Sender={envelope_mock.sender}, to={envelope_mock.sender}."
        )
    assert count + 1 == handler.unsupported_protocol_count


def test_send_decoding_error():
    """Test the send_decoding_error method."""
    handler = ErrorHandler
    envelope_mock = Mock()
    envelope_mock.protocol_id = "1"
    envelope_mock.sender = "2"
    envelope_mock.to = "3"
    count = handler.decoding_error_count
    with patch.object(_default_logger, "warning") as mock_logger:
        handler.send_decoding_error(envelope_mock, _default_logger)
        mock_logger.assert_any_call(
            f"Decoding error for envelope: {envelope_mock}. Protocol_id='{envelope_mock.protocol_id}' and message are inconsistent. Sender={envelope_mock.sender}, to={envelope_mock.sender}."
        )
    assert count + 1 == handler.decoding_error_count


def test_send_unsupported_skill_1():
    """Test the send_unsupported_protocol method."""
    handler = ErrorHandler
    envelope_mock = Mock()
    envelope_mock.protocol_id = "1"
    envelope_mock.sender = "2"
    envelope_mock.to = "3"
    envelope_mock.skill_id = None
    count = handler.unsupported_skill_count
    with patch.object(_default_logger, "warning") as mock_logger:
        handler.send_unsupported_skill(envelope_mock, _default_logger)
        mock_logger.assert_any_call(
            f"Cannot handle envelope: no active handler registered for the protocol_id='{envelope_mock.protocol_id}'. Sender={envelope_mock.sender}, to={envelope_mock.sender}."
        )
    assert count + 1 == handler.unsupported_skill_count


def test_send_unsupported_skill_2():
    """Test the send_unsupported_protocol method."""
    handler = ErrorHandler
    envelope_mock = Mock()
    envelope_mock.protocol_id = "1"
    envelope_mock.sender = "2"
    envelope_mock.to = "3"
    envelope_mock.skill_id = "4"
    count = handler.unsupported_skill_count
    with patch.object(_default_logger, "warning") as mock_logger:
        handler.send_unsupported_skill(envelope_mock, _default_logger)
        mock_logger.assert_any_call(
            f"Cannot handle envelope: no active handler registered for the protocol_id='{envelope_mock.protocol_id}' and skill_id='{envelope_mock.skill_id}'. Sender={envelope_mock.sender}, to={envelope_mock.sender}."
        )
    assert count + 1 == handler.unsupported_skill_count



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""
Check that every package has existing dependencies.

Run this script from the root of the project directory:

    python scripts/check_package_dependencies.py

"""
import pprint
import sys
from functools import partial
from itertools import chain
from pathlib import Path
from typing import Dict, List, Set

import yaml

from aea.configurations.base import PackageId, PackageType, PublicId


DEFAULT_CONFIG_FILE_PATHS = []  # type: List[Path]


class DependencyNotFound(Exception):
    """Custom exception for dependencies not found."""

    def __init__(
        self,
        configuration_file: Path,
        expected_deps: Set[PackageId],
        missing_dependencies: Set[PackageId],
        *args,
    ):
        """
        Initialize DependencyNotFound exception.

        :param configuration_file: path to the checked file.
        :param expected_dependencies: expected dependencies.
        :param missing_dependencies: missing dependencies.
        :param kwargs: super class args.
        """
        super().__init__(*args)
        self.configuration_file = configuration_file
        self.expected_dependencies = expected_deps
        self.missing_dependencies = missing_dependencies


def find_all_configuration_files():
    """Find all configuration files."""
    packages_dir = Path("packages")
    return list(chain(packages_dir.glob("*/*/*/*.yaml"), default_config_file_paths()))


def default_config_file_paths():
    """Get (generator) the default config file paths."""
    for item in DEFAULT_CONFIG_FILE_PATHS:
        yield item


def get_public_id_from_yaml(configuration_file: Path):
    """
    Get the public id from yaml.

    :param configuration_file: the path to the config yaml
    """
    data = unified_yaml_load(configuration_file)
    author = data["author"]
    # handle the case when it's a package or agent config file.
    name = data["name"] if "name" in data else data["agent_name"]
    version = data["version"]
    return PublicId(author, name, version)


def find_all_packages_ids() -> Set[PackageId]:
    """Find all packages ids."""
    package_ids: Set[PackageId] = set()
    for configuration_file in find_all_configuration_files():
        package_type = PackageType(configuration_file.parts[-3][:-1])
        package_public_id = get_public_id_from_yaml(configuration_file)
        package_id = PackageId(package_type, package_public_id)
        package_ids.add(package_id)

    return package_ids


def handle_dependency_not_found(e: DependencyNotFound):
    """Handle PackageIdNotFound errors."""
    sorted_expected = list(map(str, sorted(e.expected_dependencies)))
    sorted_missing = list(map(str, sorted(e.missing_dependencies)))
    print("=" * 50)
    print(f"Package {e.configuration_file}:")
    print(f"Expected: {pprint.pformat(sorted_expected)}")
    print(f"Missing: {pprint.pformat(sorted_missing)}")


def unified_yaml_load(configuration_file: Path) -> Dict:
    """
    Load YAML file, unified (both single- and multi-paged).

    :param configuration_file: the configuration file path.
    :return: the data.
    """
    package_type = configuration_file.parent.parent.name
    with configuration_file.open() as fp:
        if package_type != "agents":
            return yaml.safe_load(fp)
        # when it is an agent configuration file,
        # we are interested only in the first page of the YAML,
        # because the dependencies are contained only there.
        data = yaml.safe_load_all(fp)
        return list(data)[0]


def check_dependencies(configuration_file: Path, all_packages_ids: Set[PackageId]):
    """
    Check dependencies of configuration file.

    :param configuration_file: path to a package configuration file.
    :param all_packages_ids: all the package ids.
    :return: None
    """
    data = unified_yaml_load(configuration_file)

    def _add_package_type(package_type, public_id_str):
        return PackageId(package_type, PublicId.from_str(public_id_str))

    def _get_package_ids(package_type, public_ids):
        return set(map(partial(_add_package_type, package_type), public_ids))

    dependencies: Set[PackageId] = set.union(
        *[
            _get_package_ids(package_type, data.get(package_type.to_plural(), set()))
            for package_type in list(PackageType)
        ]
    )

    diff = dependencies.difference(all_packages_ids)
    if len(diff) > 0:
        raise DependencyNotFound(configuration_file, dependencies, diff)


if __name__ == "__main__":
    all_packages_ids_ = find_all_packages_ids()
    failed: bool = False
    for file in find_all_configuration_files():
        try:
            print("Processing " + str(file))
            check_dependencies(file, all_packages_ids_)
        except DependencyNotFound as e_:
            handle_dependency_not_found(e_)
            failed = True

    if failed:
        print("Failed!")
        sys.exit(1)
    else:
        print("OK!")
        sys.exit(0)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Implementation of the 'aea scaffold' subcommand."""

import os
import re
import shutil
from pathlib import Path
from typing import cast

import click
from jsonschema import ValidationError

from aea import AEA_DIR
from aea.cli.fingerprint import fingerprint_item
from aea.cli.utils.context import Context
from aea.cli.utils.decorators import check_aea_project, clean_after, pass_ctx
from aea.cli.utils.loggers import logger
from aea.cli.utils.package_utils import (
    create_symlink_packages_to_vendor,
    create_symlink_vendor_to_local,
    validate_package_name,
)
from aea.configurations.base import PublicId
from aea.configurations.constants import (  # noqa: F401  # pylint: disable=unused-import
    CONNECTION,
    CONTRACT,
    DEFAULT_AEA_CONFIG_FILE,
    DEFAULT_CONNECTION_CONFIG_FILE,
    DEFAULT_CONTRACT_CONFIG_FILE,
    DEFAULT_PROTOCOL_CONFIG_FILE,
    DEFAULT_SKILL_CONFIG_FILE,
    DEFAULT_VERSION,
    PROTOCOL,
    SCAFFOLD_PUBLIC_ID,
    SKILL,
)


@click.group()
@click.option(
    "--with-symlinks",
    is_flag=True,
    help="Add symlinks from vendor to non-vendor and packages to vendor folders.",
)
@click.pass_context
@check_aea_project
def scaffold(
    click_context: click.core.Context, with_symlinks: bool
):  # pylint: disable=unused-argument
    """Scaffold a resource for the agent."""
    ctx = cast(Context, click_context.obj)
    ctx.set_config("with_symlinks", with_symlinks)


@scaffold.command()
@click.argument("connection_name", type=str, required=True)
@pass_ctx
def connection(ctx: Context, connection_name: str) -> None:
    """Add a connection scaffolding to the configuration file and agent."""
    scaffold_item(ctx, CONNECTION, connection_name)


@scaffold.command()
@click.argument("contract_name", type=str, required=True)
@pass_ctx
def contract(ctx: Context, contract_name: str) -> None:
    """Add a contract scaffolding to the configuration file and agent."""
    scaffold_item(ctx, CONTRACT, contract_name)


@scaffold.command()
@click.argument("protocol_name", type=str, required=True)
@pass_ctx
def protocol(ctx: Context, protocol_name: str):
    """Add a protocol scaffolding to the configuration file and agent."""
    scaffold_item(ctx, PROTOCOL, protocol_name)


@scaffold.command()
@click.argument("skill_name", type=str, required=True)
@pass_ctx
def skill(ctx: Context, skill_name: str):
    """Add a skill scaffolding to the configuration file and agent."""
    scaffold_item(ctx, SKILL, skill_name)


@scaffold.command()
@pass_ctx
def decision_maker_handler(ctx: Context):
    """Add a decision maker scaffolding to the configuration file and agent."""
    _scaffold_dm_handler(ctx)


@clean_after
def scaffold_item(ctx: Context, item_type: str, item_name: str) -> None:
    """
    Add an item scaffolding to the configuration file and agent.

    :param ctx: Context object.
    :param item_type: type of item.
    :param item_name: item name.

    :return: None
    :raises ClickException: if some error occures.
    """
    validate_package_name(item_name)
    author_name = ctx.agent_config.author
    loader = getattr(ctx, f"{item_type}_loader")
    default_config_filename = globals()[f"DEFAULT_{item_type.upper()}_CONFIG_FILE"]

    item_type_plural = item_type + "s"
    existing_ids = getattr(ctx.agent_config, f"{item_type}s")
    existing_ids_only_author_and_name = map(lambda x: (x.author, x.name), existing_ids)
    # check if we already have an item with the same public id
    if (author_name, item_name) in existing_ids_only_author_and_name:
        raise click.ClickException(
            f"A {item_type} with name '{item_name}' already exists. Aborting..."
        )

    agent_name = ctx.agent_config.agent_name
    click.echo(
        f"Adding {item_type} scaffold '{item_name}' to the agent '{agent_name}'..."
    )

    # create the item folder
    Path(item_type_plural).mkdir(exist_ok=True)
    dest = os.path.join(item_type_plural, item_name)
    if os.path.exists(dest):
        raise click.ClickException(
            f"A {item_type} with this name already exists. Please choose a different name and try again."
        )

    ctx.clean_paths.append(str(dest))
    try:
        # copy the item package into the agent project.
        src = Path(os.path.join(AEA_DIR, item_type_plural, "scaffold"))
        logger.debug(f"Copying {item_type} modules. src={src} dst={dest}")
        shutil.copytree(src, dest)

        # add the item to the configurations.
        logger.debug(f"Registering the {item_type} into {DEFAULT_AEA_CONFIG_FILE}")
        new_public_id = PublicId(author_name, item_name, DEFAULT_VERSION)
        existing_ids.add(new_public_id)
        with open(os.path.join(ctx.cwd, DEFAULT_AEA_CONFIG_FILE), "w") as fp:
            ctx.agent_loader.dump(ctx.agent_config, fp)

        # ensure the name in the yaml and the name of the folder are the same
        config_filepath = Path(
            ctx.cwd, item_type_plural, item_name, default_config_filename
        )
        with config_filepath.open() as fp:
            config = loader.load(fp)
        config.name = item_name
        config.author = author_name
        with config_filepath.open("w") as fp:
            loader.dump(config, fp)

        # update 'PUBLIC_ID' variable with the right public id
        init_module = Path(dest, "__init__.py")
        init_module.write_text(
            re.sub(SCAFFOLD_PUBLIC_ID, str(new_public_id), init_module.read_text())
        )

        # fingerprint item.
        fingerprint_item(ctx, item_type, new_public_id)

        if ctx.config.get("with_symlinks", False):
            click.echo(
                "Adding symlinks from vendor to non-vendor and packages to vendor folders."
            )
            create_symlink_vendor_to_local(ctx, item_type, new_public_id)
            create_symlink_packages_to_vendor(ctx)

    except ValidationError:
        raise click.ClickException(
            f"Error when validating the {item_type} configuration file."
        )
    except Exception as e:
        raise click.ClickException(str(e))


def _scaffold_dm_handler(ctx: Context):
    """Add a scaffolded decision maker handler to the project and configuration."""
    existing_dm_handler = ctx.agent_config.decision_maker_handler

    # check if we already have a decision maker in the project
    if existing_dm_handler != {}:
        raise click.ClickException(
            "A decision maker handler specification already exists. Aborting..."
        )

    dest = Path("decision_maker.py")
    agent_name = ctx.agent_config.agent_name
    click.echo("Adding decision maker scaffold to the agent '{}'...".format(agent_name))

    # create the file name
    dotted_path = ".decision_maker::DecisionMakerHandler"
    try:
        # copy the item package into the agent project.
        src = Path(os.path.join(AEA_DIR, "decision_maker", "scaffold.py"))
        logger.debug("Copying decision maker. src={} dst={}".format(src, dest))
        shutil.copyfile(src, dest)

        # add the item to the configurations.
        logger.debug(
            "Registering the decision_maker into {}".format(DEFAULT_AEA_CONFIG_FILE)
        )
        ctx.agent_config.decision_maker_handler = {
            "dotted_path": str(dotted_path),
            "file_path": str(os.path.join(".", dest)),
        }
        ctx.agent_loader.dump(
            ctx.agent_config, open(os.path.join(ctx.cwd, DEFAULT_AEA_CONFIG_FILE), "w")
        )

    except Exception as e:
        os.remove(dest)
        raise click.ClickException(str(e))



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This test module contains the tests for the `aea launch` sub-command."""
import json
import os
import sys
import uuid

import pytest

from aea.test_tools.test_cases import AEATestCaseMany

from tests.common.pexpect_popen import PexpectWrapper


class TestLaunchEndToEnd(AEATestCaseMany):
    """Perform aea launch end to end test."""

    key = "seller_service"
    value = None

    registration_agent_connection = {
        "delegate_uri": "127.0.0.1:11011",
        "entry_peers": [],
        "ledger_id": "fetchai",
        "local_uri": "127.0.0.1:9011",
        "log_file": "libp2p_node.log",
        "public_uri": "127.0.0.1:9011",
    }

    search_agent_connection = {
        "delegate_uri": "127.0.0.1:11012",
        "entry_peers": [],
        "ledger_id": "fetchai",
        "local_uri": "127.0.0.1:9012",
        "log_file": "libp2p_node.log",
        "public_uri": "127.0.0.1:9012",
    }

    @pytest.mark.integration
    def test_end_to_end(self):
        """Perform end to end test with simple register/search agents."""
        registration_agent_name = "registration_agent"
        self.value = uuid.uuid4().hex
        self.fetch_agent(
            "fetchai/simple_service_registration",
            agent_name=registration_agent_name,
            is_local=True,
        )

        self.run_cli_command(
            "config",
            "set",
            "vendor.fetchai.connections.p2p_libp2p.config",
            "--type",
            "dict",
            json.dumps(self.registration_agent_connection),
            cwd=registration_agent_name,
        )
        self.run_cli_command(
            "config",
            "set",
            "vendor.fetchai.skills.simple_service_registration.models.strategy.args.service_data",
            "--type",
            "dict",
            json.dumps({"key": self.key, "value": self.value}),
            cwd=registration_agent_name,
        )

        search_agent_name = "search_agent"
        self.fetch_agent(
            "fetchai/simple_service_search",
            agent_name=search_agent_name,
            is_local=True,
        )
        self.run_cli_command(
            "config",
            "set",
            "vendor.fetchai.connections.p2p_libp2p.config",
            "--type",
            "dict",
            json.dumps(self.search_agent_connection),
            cwd=search_agent_name,
        )
        self.run_cli_command(
            "config",
            "set",
            "vendor.fetchai.skills.simple_service_search.models.strategy.args.search_query",
            "--type",
            "dict",
            json.dumps(
                {
                    "constraint_type": "==",
                    "search_key": self.key,
                    "search_value": self.value,
                }
            ),
            cwd=search_agent_name,
        )
        self.run_cli_command(
            "config",
            "set",
            "vendor.fetchai.skills.simple_service_search.behaviours.service_search.args.tick_interval",
            "--type",
            "int",
            "2",
            cwd=search_agent_name,
        )

        proc = PexpectWrapper(  # nosec
            [
                sys.executable,
                "-m",
                "aea.cli",
                "-v",
                "DEBUG",
                "launch",
                registration_agent_name,
                search_agent_name,
            ],
            env=os.environ,
            maxread=10000,
            encoding="utf-8",
            logfile=sys.stdout,
        )
        try:
            proc.expect_all(
                [f"[{search_agent_name}] found number of agents=1, search_response"],
                timeout=30,
            )
        finally:
            proc.control_c()
            proc.expect("Exit cli. code: 0", timeout=30)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains tests for transaction."""
from typing import Type
from unittest.mock import patch

import pytest

from aea.common import Address
from aea.protocols.base import Message
from aea.protocols.dialogue.base import Dialogue as BaseDialogue
from aea.protocols.dialogue.base import DialogueLabel

from packages.fetchai.protocols.state_update.dialogues import (
    StateUpdateDialogue as BaseStateUpdateDialogue,
)
from packages.fetchai.protocols.state_update.dialogues import (
    StateUpdateDialogues as BaseStateUpdateDialogues,
)
from packages.fetchai.protocols.state_update.message import (
    StateUpdateMessage,
    _default_logger,
)


class TestStateUpdateMessage:
    """Test the StateUpdateMessage."""

    def test_message_consistency(self):
        """Test for an error in consistency of a message."""
        currency_endowment = {"FET": 100}
        good_endowment = {"a_good": 2}
        exchange_params = {"FET": 10.0}
        utility_params = {"a_good": 20.0}
        assert StateUpdateMessage(
            performative=StateUpdateMessage.Performative.INITIALIZE,
            amount_by_currency_id=currency_endowment,
            quantities_by_good_id=good_endowment,
            exchange_params_by_currency_id=exchange_params,
            utility_params_by_good_id=utility_params,
        )
        currency_change = {"FET": 10}
        good_change = {"a_good": 1}
        stum = StateUpdateMessage(
            performative=StateUpdateMessage.Performative.APPLY,
            amount_by_currency_id=currency_change,
            quantities_by_good_id=good_change,
        )
        assert stum._is_consistent()
        assert len(stum.valid_performatives) == 2

    def test_message_inconsistency(self):
        """Test for an error in consistency of a message."""
        currency_endowment = {"FET": 100}
        good_endowment = {"a_good": 2}
        exchange_params = {"UNKNOWN": 10.0}
        utility_params = {"a_good": 20.0}
        with pytest.raises(ValueError, match="Field .* is not supported"):
            StateUpdateMessage(
                performative=StateUpdateMessage.Performative.INITIALIZE,
                amount_by_currency_id=currency_endowment,
                quantities_by_good_id=good_endowment,
                exchange_params_by_currency_id=exchange_params,
                utility_params_by_good_id=utility_params,
                non_exists_field="some value",
            )


class TestSerialization:
    """Test state update message serialization."""

    def test_serialization_initialize(self):
        """Test serialization of initialize message."""
        currency_endowment = {"FET": 100}
        good_endowment = {"a_good": 2}
        exchange_params = {"FET": 10.0}
        utility_params = {"a_good": 20.0}
        msg = StateUpdateMessage(
            performative=StateUpdateMessage.Performative.INITIALIZE,
            amount_by_currency_id=currency_endowment,
            quantities_by_good_id=good_endowment,
            exchange_params_by_currency_id=exchange_params,
            utility_params_by_good_id=utility_params,
        )
        encoded_msg = msg.serializer.encode(msg)
        decoded_msg = msg.serializer.decode(encoded_msg)
        assert msg == decoded_msg

    def test_serialization_apply(self):
        """Test serialization of apply message."""
        currency_change = {"FET": 10}
        good_change = {"a_good": 1}
        msg = StateUpdateMessage(
            performative=StateUpdateMessage.Performative.APPLY,
            amount_by_currency_id=currency_change,
            quantities_by_good_id=good_change,
        )
        assert msg._is_consistent()
        assert len(msg.valid_performatives) == 2
        encoded_msg = msg.serializer.encode(msg)
        decoded_msg = msg.serializer.decode(encoded_msg)
        assert msg == decoded_msg


def test_serialization_negative():
    """Test serialization when performative is not recognized."""
    currency_change = {"FET": 10}
    good_change = {"a_good": 1}
    msg = StateUpdateMessage(
        performative=StateUpdateMessage.Performative.APPLY,
        amount_by_currency_id=currency_change,
        quantities_by_good_id=good_change,
    )

    with patch.object(StateUpdateMessage.Performative, "__eq__", return_value=False):
        with pytest.raises(
            ValueError, match=f"Performative not valid: {msg.performative}"
        ):
            msg.serializer.encode(msg)

    encoded_tx_bytes = msg.serializer.encode(msg)
    with patch.object(StateUpdateMessage.Performative, "__eq__", return_value=False):
        with pytest.raises(
            ValueError, match=f"Performative not valid: {msg.performative}"
        ):
            msg.serializer.decode(encoded_tx_bytes)


def test_performative_str():
    """Test performative __str__."""
    assert str(StateUpdateMessage.Performative.INITIALIZE) == "initialize"
    assert str(StateUpdateMessage.Performative.APPLY) == "apply"


def test_light_protocol_rule_3_target_less_than_message_id():
    """Test that if message_id is not 1, target must be > message_id"""
    with patch.object(_default_logger, "error") as mock_logger:
        currency_endowment = {"FET": 100}
        good_endowment = {"a_good": 2}
        exchange_params = {"FET": 10.0}
        utility_params = {"a_good": 20.0}
        message_id = 2
        target = 2
        assert StateUpdateMessage(
            message_id=message_id,
            target=target,
            performative=StateUpdateMessage.Performative.INITIALIZE,
            amount_by_currency_id=currency_endowment,
            quantities_by_good_id=good_endowment,
            exchange_params_by_currency_id=exchange_params,
            utility_params_by_good_id=utility_params,
        )

        mock_logger.assert_any_call(
            f"Invalid 'target'. Expected an integer between 1 and {message_id - 1} inclusive. Found {target}."
        )


def test_dialogues():
    """Test intiaontiation of dialogues."""
    state_update_dialogues = StateUpdateDialogues("agent_addr")
    msg, dialogue = state_update_dialogues.create(
        counterparty="abc",
        performative=StateUpdateMessage.Performative.INITIALIZE,
        amount_by_currency_id={},
        quantities_by_good_id={},
        exchange_params_by_currency_id={},
        utility_params_by_good_id={},
    )
    assert dialogue is not None


class StateUpdateDialogue(BaseStateUpdateDialogue):
    """The dialogue class maintains state of a dialogue and manages it."""

    def __init__(
        self,
        dialogue_label: DialogueLabel,
        self_address: Address,
        role: BaseDialogue.Role,
        message_class: Type[StateUpdateMessage],
    ) -> None:
        """
        Initialize a dialogue.

        :param dialogue_label: the identifier of the dialogue
        :param self_address: the address of the entity for whom this dialogue is maintained
        :param role: the role of the agent this dialogue is maintained for

        :return: None
        """
        BaseStateUpdateDialogue.__init__(
            self,
            dialogue_label=dialogue_label,
            self_address=self_address,
            role=role,
            message_class=message_class,
        )


class StateUpdateDialogues(BaseStateUpdateDialogues):
    """The dialogues class keeps track of all dialogues."""

    def __init__(self, self_address: Address) -> None:
        """
        Initialize dialogues.

        :return: None
        """

        def role_from_first_message(  # pylint: disable=unused-argument
            message: Message, receiver_address: Address
        ) -> BaseDialogue.Role:
            """Infer the role of the agent from an incoming/outgoing first message

            :param message: an incoming/outgoing first message
            :param receiver_address: the address of the receiving agent
            :return: The role of the agent
            """
            return StateUpdateDialogue.Role.SKILL

        BaseStateUpdateDialogues.__init__(
            self,
            self_address=self_address,
            role_from_first_message=role_from_first_message,
            dialogue_class=StateUpdateDialogue,
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2020 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests for the sym link module."""

import os
import shutil
import tempfile
from pathlib import Path

from aea.helpers.sym_link import create_symlink


def test_create_symlink():
    """Test create_symlink method."""
    t = Path(tempfile.mkdtemp())
    cwd = os.getcwd()
    os.chdir(t)
    try:
        link = Path(os.path.join(t, "here"))
        target = Path(os.path.join(t, "test", "nested"))
        os.makedirs(target)
        create_symlink(link, target, t)
        assert os.path.islink(link)
        assert os.readlink("here") == os.path.join("test", "nested")
    finally:
        os.chdir(cwd)
        shutil.rmtree(t)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2020 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests for the 'helpers/file_io' module."""
import os
import tempfile
from pathlib import Path
from unittest.mock import patch

import pytest

import aea
from aea.configurations.base import PublicId
from aea.helpers.file_io import _decode, envelope_from_bytes, lock_file, write_envelope
from aea.mail.base import Envelope


class TestFileLock:
    """Test for filelocks."""

    def test_lock_file_ok(self):
        """Work ok ok for random file."""
        with tempfile.TemporaryFile() as fp:
            with lock_file(fp):
                pass

    def test_lock_file_error(self):
        """Fail on closed file."""
        with tempfile.TemporaryFile() as fp:
            fp.close()
            with pytest.raises(ValueError):
                with lock_file(fp):
                    pass


def test_envelope_serialization():
    """Test envelope serialization/deserialization with files."""
    envelope = Envelope(
        to="to",
        sender="sender",
        protocol_id=PublicId("author", "name", "0.1.0"),
        message=b"",
    )
    with tempfile.TemporaryDirectory() as temp_dir:
        output_file = Path(os.path.join(temp_dir, "output_file"))
        with output_file.open(mode="wb") as fout:
            write_envelope(envelope, fout)

        actual_envelope = envelope_from_bytes(output_file.read_bytes())

    assert envelope == actual_envelope


def test_decode_fails():
    """Test decode fails."""
    with pytest.raises(
        ValueError,
        match="Expected at least 5 values separated by commas and last value being empty or new line, got 1",
    ):
        _decode(b"")


def test_envelope_from_bytes_bad_format():
    """Test envelope_from_bytes, when the input has a bad format we raise ValueError."""
    test_error_message = "Something bad happened."
    _bytes = b""
    with patch(
        "aea.helpers.file_io._decode", side_effect=ValueError(test_error_message)
    ):
        with patch.object(aea.helpers.file_io._default_logger, "error") as mock_error:
            envelope_from_bytes(_bytes)
            mock_error.assert_called_with(
                f"Bad formatted input: {_bytes}. {test_error_message}"
            )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2020 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Implementation of the 'aea eject' subcommand."""
import shutil
from pathlib import Path
from typing import cast

import click

from aea.cli.remove import ItemRemoveHelper
from aea.cli.utils.click_utils import PublicIdParameter
from aea.cli.utils.config import (
    set_cli_author,
    try_to_load_agent_config,
    update_item_config,
)
from aea.cli.utils.context import Context
from aea.cli.utils.decorators import check_aea_project, clean_after, pass_ctx
from aea.cli.utils.package_utils import (
    copy_package_directory,
    create_symlink_packages_to_vendor,
    create_symlink_vendor_to_local,
    fingerprint_all,
    get_package_path,
    is_item_present,
    replace_all_import_statements,
    update_item_public_id_in_init,
    update_references,
)
from aea.configurations.base import (
    ComponentId,
    ComponentType,
    PackageId,
    PackageType,
    PublicId,
)
from aea.configurations.constants import (
    CONNECTION,
    CONTRACT,
    DEFAULT_VERSION,
    PROTOCOL,
    SKILL,
)
from aea.configurations.utils import get_latest_component_id_from_prefix
from aea.helpers.base import find_topological_order, reachable_nodes


@click.group()
@click.option(
    "--with-symlinks",
    is_flag=True,
    help="Add symlinks from vendor to non-vendor and packages to vendor folders.",
)
@click.option(
    "-q",
    "--quiet",
    "quiet",
    is_flag=True,
    required=False,
    default=False,
    help="If provided, the command will not ask the user for confirmation.",
)
@click.pass_context
@check_aea_project
def eject(click_context: click.core.Context, quiet: bool, with_symlinks: bool):
    """Eject an installed item."""
    click_context.obj.set_config("quiet", quiet)
    click_context.obj.set_config("with_symlinks", with_symlinks)
    set_cli_author(click_context)


@eject.command()
@click.argument("public_id", type=PublicIdParameter(), required=True)
@pass_ctx
def connection(ctx: Context, public_id: PublicId):
    """Eject an installed connection."""
    quiet = ctx.config.get("quiet")
    with_symlinks = ctx.config.get("with_symlinks")
    _eject_item(ctx, CONNECTION, public_id, quiet=quiet, with_symlinks=with_symlinks)


@eject.command()
@click.argument("public_id", type=PublicIdParameter(), required=True)
@pass_ctx
def contract(ctx: Context, public_id: PublicId):
    """Eject an installed contract."""
    quiet = ctx.config.get("quiet")
    with_symlinks = ctx.config.get("with_symlinks")
    _eject_item(ctx, CONTRACT, public_id, quiet=quiet, with_symlinks=with_symlinks)


@eject.command()
@click.argument("public_id", type=PublicIdParameter(), required=True)
@pass_ctx
def protocol(ctx: Context, public_id: PublicId):
    """Eject an installed protocol."""
    quiet = ctx.config.get("quiet")
    with_symlinks = ctx.config.get("with_symlinks")
    _eject_item(ctx, PROTOCOL, public_id, quiet=quiet, with_symlinks=with_symlinks)


@eject.command()
@click.argument("public_id", type=PublicIdParameter(), required=True)
@pass_ctx
def skill(ctx: Context, public_id: PublicId):
    """Eject an installed skill."""
    quiet = ctx.config.get("quiet")
    with_symlinks = ctx.config.get("with_symlinks")
    _eject_item(ctx, SKILL, public_id, quiet=quiet, with_symlinks=with_symlinks)


@clean_after
def _eject_item(
    ctx: Context,
    item_type: str,
    public_id: PublicId,
    quiet: bool = True,
    with_symlinks: bool = False,
):
    """
    Eject item from installed (vendor) to custom folder.

    :param ctx: context object.
    :param item_type: item type.
    :param public_id: item public ID.
    :param quiet: if false, the function will ask the user in case of recursive eject.

    :return: None
    :raises: ClickException if item is absent at source path or present at destenation path.
    """
    # we know cli_author is set because of the above checks.
    cli_author: str = cast(str, ctx.config.get("cli_author"))
    item_type_plural = item_type + "s"
    if not is_item_present(
        ctx, item_type, public_id, is_vendor=True, with_version=True
    ):  # pragma: no cover
        raise click.ClickException(
            f"{item_type.title()} {public_id} not found in agent's vendor items."
        )
    src = get_package_path(ctx, item_type, public_id)
    dst = get_package_path(ctx, item_type, public_id, is_vendor=False)
    if is_item_present(ctx, item_type, public_id, is_vendor=False):  # pragma: no cover
        raise click.ClickException(
            f"{item_type.title()} {public_id} is already a non-vendor package."
        )

    if public_id.package_version.is_latest:
        # get 'concrete' public id, in case it is 'latest'
        component_prefix = ComponentType(item_type), public_id.author, public_id.name
        component_id = get_latest_component_id_from_prefix(
            ctx.agent_config, component_prefix
        )
        # component id is necessarily found, due to the checks above.
        public_id = cast(ComponentId, component_id).public_id

    package_id = PackageId(PackageType(item_type), public_id)

    click.echo(
        f"Ejecting item {package_id.package_type.value} {str(package_id.public_id)}"
    )

    # first, eject all the vendor packages that depend on this
    item_remover = ItemRemoveHelper(ctx, ignore_non_vendor=True)
    reverse_dependencies = (
        item_remover.get_agent_dependencies_with_reverse_dependencies()
    )
    reverse_reachable_dependencies = reachable_nodes(reverse_dependencies, {package_id})
    # the reversed topological order of a graph
    # is the topological order of the reverse graph.
    eject_order = list(reversed(find_topological_order(reverse_reachable_dependencies)))
    eject_order.remove(package_id)
    if len(eject_order) > 0 and not quiet:
        click.echo(f"The following vendor packages will be ejected: {eject_order}")
        answer = click.confirm("Do you want to proceed?")
        if not answer:
            click.echo("Aborted.")
            return

    for dependency_package_id in eject_order:
        # 'dependency_package_id' depends on 'package_id',
        # so we need to eject it first
        _eject_item(
            ctx,
            dependency_package_id.package_type.value,
            dependency_package_id.public_id,
            quiet=True,
        )

    # copy the vendor package into the non-vendor packages
    ctx.clean_paths.append(dst)
    copy_package_directory(Path(src), dst)

    new_public_id = PublicId(cli_author, public_id.name, DEFAULT_VERSION)
    update_item_config(
        item_type, Path(dst), author=new_public_id.author, version=new_public_id.version
    )
    update_item_public_id_in_init(item_type, Path(dst), new_public_id)
    shutil.rmtree(src)

    # update references in all the other packages
    component_type = ComponentType(item_type_plural[:-1])
    old_component_id = ComponentId(component_type, public_id)
    new_component_id = ComponentId(component_type, new_public_id)
    update_references(ctx, {old_component_id: new_component_id})

    # need to reload agent configuration with the updated references
    try_to_load_agent_config(ctx)

    # replace import statements in all the non-vendor packages
    replace_all_import_statements(
        Path(ctx.cwd), ComponentType(item_type), public_id, new_public_id
    )

    # fingerprint all (non-vendor) packages
    fingerprint_all(ctx)

    if with_symlinks:
        click.echo(
            "Adding symlinks from vendor to non-vendor and packages to vendor folders."
        )
        create_symlink_vendor_to_local(ctx, item_type, new_public_id)
        create_symlink_packages_to_vendor(ctx)

    click.echo(
        f"Successfully ejected {item_type} {public_id} to {dst} as {new_public_id}."
    )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""
This module contains the classes required for dialogue management.

- OefSearchDialogue: The dialogue class maintains state of a dialogue of type oef search and manages it.
- OefSearchDialogues: The dialogues class keeps track of all dialogues of type oef search.
"""

from aea.protocols.base import Address, Message
from aea.protocols.dialogue.base import Dialogue as BaseDialogue
from aea.skills.base import Model

from packages.fetchai.protocols.default.dialogues import (
    DefaultDialogue as BaseDefaultDialogue,
)
from packages.fetchai.protocols.default.dialogues import (
    DefaultDialogues as BaseDefaultDialogues,
)
from packages.fetchai.protocols.http.dialogues import HttpDialogue as BaseHttpDialogue
from packages.fetchai.protocols.http.dialogues import HttpDialogues as BaseHttpDialogues
from packages.fetchai.protocols.oef_search.dialogues import (
    OefSearchDialogue as BaseOefSearchDialogue,
)
from packages.fetchai.protocols.oef_search.dialogues import (
    OefSearchDialogues as BaseOefSearchDialogues,
)


DefaultDialogue = BaseDefaultDialogue


class DefaultDialogues(Model, BaseDefaultDialogues):
    """The dialogues class keeps track of all dialogues."""

    def __init__(self, **kwargs) -> None:
        """
        Initialize dialogues.

        :return: None
        """
        Model.__init__(self, **kwargs)

        def role_from_first_message(  # pylint: disable=unused-argument
            message: Message, receiver_address: Address
        ) -> BaseDialogue.Role:
            """Infer the role of the agent from an incoming/outgoing first message

            :param message: an incoming/outgoing first message
            :param receiver_address: the address of the receiving agent
            :return: The role of the agent
            """
            return DefaultDialogue.Role.AGENT

        BaseDefaultDialogues.__init__(
            self,
            self_address=self.context.agent_address,
            role_from_first_message=role_from_first_message,
        )


HttpDialogue = BaseHttpDialogue


class HttpDialogues(Model, BaseHttpDialogues):
    """This class keeps track of all http dialogues."""

    def __init__(self, **kwargs) -> None:
        """
        Initialize dialogues.

        :param agent_address: the address of the agent for whom dialogues are maintained
        :return: None
        """
        Model.__init__(self, **kwargs)

        def role_from_first_message(  # pylint: disable=unused-argument
            message: Message, receiver_address: Address
        ) -> BaseDialogue.Role:
            """Infer the role of the agent from an incoming/outgoing first message

            :param message: an incoming/outgoing first message
            :param receiver_address: the address of the receiving agent
            :return: The role of the agent
            """
            return BaseHttpDialogue.Role.CLIENT

        BaseHttpDialogues.__init__(
            self,
            self_address=self.context.agent_address,
            role_from_first_message=role_from_first_message,
        )


OefSearchDialogue = BaseOefSearchDialogue


class OefSearchDialogues(Model, BaseOefSearchDialogues):
    """This class keeps track of all oef_search dialogues."""

    def __init__(self, **kwargs) -> None:
        """
        Initialize dialogues.

        :param agent_address: the address of the agent for whom dialogues are maintained
        :return: None
        """
        Model.__init__(self, **kwargs)

        def role_from_first_message(  # pylint: disable=unused-argument
            message: Message, receiver_address: Address
        ) -> BaseDialogue.Role:
            """Infer the role of the agent from an incoming/outgoing first message

            :param message: an incoming/outgoing first message
            :param receiver_address: the address of the receiving agent
            :return: The role of the agent
            """
            return BaseOefSearchDialogue.Role.AGENT

        BaseOefSearchDialogues.__init__(
            self,
            self_address=self.context.agent_address,
            role_from_first_message=role_from_first_message,
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2020 fetchai
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains oef_search's message definition."""

import logging
from typing import Set, Tuple, cast

from aea.configurations.base import PublicId
from aea.exceptions import AEAEnforceError, enforce
from aea.protocols.base import Message

from packages.fetchai.protocols.oef_search.custom_types import (
    AgentsInfo as CustomAgentsInfo,
)
from packages.fetchai.protocols.oef_search.custom_types import (
    Description as CustomDescription,
)
from packages.fetchai.protocols.oef_search.custom_types import (
    OefErrorOperation as CustomOefErrorOperation,
)
from packages.fetchai.protocols.oef_search.custom_types import Query as CustomQuery


_default_logger = logging.getLogger("aea.packages.fetchai.protocols.oef_search.message")

DEFAULT_BODY_SIZE = 4


class OefSearchMessage(Message):
    """A protocol for interacting with an OEF search service."""

    protocol_id = PublicId.from_str("fetchai/oef_search:0.10.0")

    AgentsInfo = CustomAgentsInfo

    Description = CustomDescription

    OefErrorOperation = CustomOefErrorOperation

    Query = CustomQuery

    class Performative(Message.Performative):
        """Performatives for the oef_search protocol."""

        OEF_ERROR = "oef_error"
        REGISTER_SERVICE = "register_service"
        SEARCH_RESULT = "search_result"
        SEARCH_SERVICES = "search_services"
        SUCCESS = "success"
        UNREGISTER_SERVICE = "unregister_service"

        def __str__(self):
            """Get the string representation."""
            return str(self.value)

    _performatives = {
        "oef_error",
        "register_service",
        "search_result",
        "search_services",
        "success",
        "unregister_service",
    }

    class _SlotsCls:
        __slots__ = (
            "agents",
            "agents_info",
            "dialogue_reference",
            "message_id",
            "oef_error_operation",
            "performative",
            "query",
            "service_description",
            "target",
        )

    def __init__(
        self,
        performative: Performative,
        dialogue_reference: Tuple[str, str] = ("", ""),
        message_id: int = 1,
        target: int = 0,
        **kwargs,
    ):
        """
        Initialise an instance of OefSearchMessage.

        :param message_id: the message id.
        :param dialogue_reference: the dialogue reference.
        :param target: the message target.
        :param performative: the message performative.
        """
        super().__init__(
            dialogue_reference=dialogue_reference,
            message_id=message_id,
            target=target,
            performative=OefSearchMessage.Performative(performative),
            **kwargs,
        )

    @property
    def valid_performatives(self) -> Set[str]:
        """Get valid performatives."""
        return self._performatives

    @property
    def dialogue_reference(self) -> Tuple[str, str]:
        """Get the dialogue_reference of the message."""
        enforce(self.is_set("dialogue_reference"), "dialogue_reference is not set.")
        return cast(Tuple[str, str], self.get("dialogue_reference"))

    @property
    def message_id(self) -> int:
        """Get the message_id of the message."""
        enforce(self.is_set("message_id"), "message_id is not set.")
        return cast(int, self.get("message_id"))

    @property
    def performative(self) -> Performative:  # type: ignore # noqa: F821
        """Get the performative of the message."""
        enforce(self.is_set("performative"), "performative is not set.")
        return cast(OefSearchMessage.Performative, self.get("performative"))

    @property
    def target(self) -> int:
        """Get the target of the message."""
        enforce(self.is_set("target"), "target is not set.")
        return cast(int, self.get("target"))

    @property
    def agents(self) -> Tuple[str, ...]:
        """Get the 'agents' content from the message."""
        enforce(self.is_set("agents"), "'agents' content is not set.")
        return cast(Tuple[str, ...], self.get("agents"))

    @property
    def agents_info(self) -> CustomAgentsInfo:
        """Get the 'agents_info' content from the message."""
        enforce(self.is_set("agents_info"), "'agents_info' content is not set.")
        return cast(CustomAgentsInfo, self.get("agents_info"))

    @property
    def oef_error_operation(self) -> CustomOefErrorOperation:
        """Get the 'oef_error_operation' content from the message."""
        enforce(
            self.is_set("oef_error_operation"),
            "'oef_error_operation' content is not set.",
        )
        return cast(CustomOefErrorOperation, self.get("oef_error_operation"))

    @property
    def query(self) -> CustomQuery:
        """Get the 'query' content from the message."""
        enforce(self.is_set("query"), "'query' content is not set.")
        return cast(CustomQuery, self.get("query"))

    @property
    def service_description(self) -> CustomDescription:
        """Get the 'service_description' content from the message."""
        enforce(
            self.is_set("service_description"),
            "'service_description' content is not set.",
        )
        return cast(CustomDescription, self.get("service_description"))

    def _is_consistent(self) -> bool:
        """Check that the message follows the oef_search protocol."""
        try:
            enforce(
                type(self.dialogue_reference) == tuple,
                "Invalid type for 'dialogue_reference'. Expected 'tuple'. Found '{}'.".format(
                    type(self.dialogue_reference)
                ),
            )
            enforce(
                type(self.dialogue_reference[0]) == str,
                "Invalid type for 'dialogue_reference[0]'. Expected 'str'. Found '{}'.".format(
                    type(self.dialogue_reference[0])
                ),
            )
            enforce(
                type(self.dialogue_reference[1]) == str,
                "Invalid type for 'dialogue_reference[1]'. Expected 'str'. Found '{}'.".format(
                    type(self.dialogue_reference[1])
                ),
            )
            enforce(
                type(self.message_id) == int,
                "Invalid type for 'message_id'. Expected 'int'. Found '{}'.".format(
                    type(self.message_id)
                ),
            )
            enforce(
                type(self.target) == int,
                "Invalid type for 'target'. Expected 'int'. Found '{}'.".format(
                    type(self.target)
                ),
            )

            # Light Protocol Rule 2
            # Check correct performative
            enforce(
                type(self.performative) == OefSearchMessage.Performative,
                "Invalid 'performative'. Expected either of '{}'. Found '{}'.".format(
                    self.valid_performatives, self.performative
                ),
            )

            # Check correct contents
            actual_nb_of_contents = len(self._body) - DEFAULT_BODY_SIZE
            expected_nb_of_contents = 0
            if self.performative == OefSearchMessage.Performative.REGISTER_SERVICE:
                expected_nb_of_contents = 1
                enforce(
                    type(self.service_description) == CustomDescription,
                    "Invalid type for content 'service_description'. Expected 'Description'. Found '{}'.".format(
                        type(self.service_description)
                    ),
                )
            elif self.performative == OefSearchMessage.Performative.UNREGISTER_SERVICE:
                expected_nb_of_contents = 1
                enforce(
                    type(self.service_description) == CustomDescription,
                    "Invalid type for content 'service_description'. Expected 'Description'. Found '{}'.".format(
                        type(self.service_description)
                    ),
                )
            elif self.performative == OefSearchMessage.Performative.SEARCH_SERVICES:
                expected_nb_of_contents = 1
                enforce(
                    type(self.query) == CustomQuery,
                    "Invalid type for content 'query'. Expected 'Query'. Found '{}'.".format(
                        type(self.query)
                    ),
                )
            elif self.performative == OefSearchMessage.Performative.SEARCH_RESULT:
                expected_nb_of_contents = 2
                enforce(
                    type(self.agents) == tuple,
                    "Invalid type for content 'agents'. Expected 'tuple'. Found '{}'.".format(
                        type(self.agents)
                    ),
                )
                enforce(
                    all(type(element) == str for element in self.agents),
                    "Invalid type for tuple elements in content 'agents'. Expected 'str'.",
                )
                enforce(
                    type(self.agents_info) == CustomAgentsInfo,
                    "Invalid type for content 'agents_info'. Expected 'AgentsInfo'. Found '{}'.".format(
                        type(self.agents_info)
                    ),
                )
            elif self.performative == OefSearchMessage.Performative.SUCCESS:
                expected_nb_of_contents = 1
                enforce(
                    type(self.agents_info) == CustomAgentsInfo,
                    "Invalid type for content 'agents_info'. Expected 'AgentsInfo'. Found '{}'.".format(
                        type(self.agents_info)
                    ),
                )
            elif self.performative == OefSearchMessage.Performative.OEF_ERROR:
                expected_nb_of_contents = 1
                enforce(
                    type(self.oef_error_operation) == CustomOefErrorOperation,
                    "Invalid type for content 'oef_error_operation'. Expected 'OefErrorOperation'. Found '{}'.".format(
                        type(self.oef_error_operation)
                    ),
                )

            # Check correct content count
            enforce(
                expected_nb_of_contents == actual_nb_of_contents,
                "Incorrect number of contents. Expected {}. Found {}".format(
                    expected_nb_of_contents, actual_nb_of_contents
                ),
            )

            # Light Protocol Rule 3
            if self.message_id == 1:
                enforce(
                    self.target == 0,
                    "Invalid 'target'. Expected 0 (because 'message_id' is 1). Found {}.".format(
                        self.target
                    ),
                )
            else:
                enforce(
                    0 < self.target < self.message_id,
                    "Invalid 'target'. Expected an integer between 1 and {} inclusive. Found {}.".format(
                        self.message_id - 1, self.target,
                    ),
                )
        except (AEAEnforceError, ValueError, KeyError) as e:
            _default_logger.error(str(e))
            return False

        return True



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2020 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains tests for aea.contracts.base."""

import logging
import os
from pathlib import Path
from typing import cast
from unittest.mock import patch

import pytest
import web3

from aea.configurations.base import ComponentType, ContractConfig
from aea.configurations.loader import load_component_configuration
from aea.contracts import contract_registry
from aea.contracts.base import Contract
from aea.contracts.scaffold.contract import MyScaffoldContract
from aea.crypto.ethereum import DEFAULT_ADDRESS as ETHEREUM_DEFAULT_ADDRESS
from aea.crypto.fetchai import DEFAULT_ADDRESS as FETCHAI_DEFAULT_ADDRESS
from aea.crypto.registries import crypto_registry, ledger_apis_registry
from aea.exceptions import AEAComponentLoadException

from tests.conftest import ETHEREUM, FETCHAI, ROOT_DIR, make_uri


logger = logging.getLogger(__name__)


def test_from_dir():
    """Tests the from dir and from config methods."""
    contract = Contract.from_dir(
        os.path.join(ROOT_DIR, "tests", "data", "dummy_contract")
    )
    assert contract is not None
    assert contract.contract_interface is not None
    assert isinstance(contract.contract_interface, dict)


def test_from_config_and_registration():
    """Tests the from config method and contract registry registration."""

    directory = Path(ROOT_DIR, "tests", "data", "dummy_contract")
    configuration = load_component_configuration(ComponentType.CONTRACT, directory)
    configuration._directory = directory
    configuration = cast(ContractConfig, configuration)

    if str(configuration.public_id) in contract_registry.specs:
        contract_registry.specs.pop(str(configuration.public_id))

    contract = Contract.from_config(configuration)
    assert contract is not None
    assert contract.contract_interface is not None
    assert isinstance(contract.contract_interface, dict)
    assert contract.configuration == configuration
    assert contract.id == configuration.public_id

    # the contract is registered as side-effect
    assert str(contract.public_id) in contract_registry.specs

    try:
        contract_registry.specs.pop(str(configuration.public_id))
    except Exception as e:
        logger.exception(e)


def test_from_config_negative():
    """Tests the from config method raises."""

    directory = Path(ROOT_DIR, "tests", "data", "dummy_contract")
    configuration = load_component_configuration(ComponentType.CONTRACT, directory)
    configuration._directory = directory
    configuration = cast(ContractConfig, configuration)

    if str(configuration.public_id) in contract_registry.specs:
        contract_registry.specs.pop(str(configuration.public_id))

    configuration.class_name = "WrongName"
    with pytest.raises(AEAComponentLoadException):
        _ = Contract.from_config(configuration)

    try:
        contract_registry.specs.pop(str(configuration.public_id))
    except Exception as e:
        logger.exception(e)


def test_non_implemented_class_methods():
    """Tests the non implemented class methods."""
    with pytest.raises(NotImplementedError):
        Contract.get_raw_transaction("ledger_api", "contract_address")

    with pytest.raises(NotImplementedError):
        Contract.get_raw_message("ledger_api", "contract_address")

    with pytest.raises(NotImplementedError):
        Contract.get_state("ledger_api", "contract_address")


@pytest.fixture()
def dummy_contract(request):
    """Dummy contract fixture."""
    directory = Path(ROOT_DIR, "tests", "data", "dummy_contract")
    configuration = load_component_configuration(ComponentType.CONTRACT, directory)
    configuration._directory = directory
    configuration = cast(ContractConfig, configuration)

    if str(configuration.public_id) in contract_registry.specs:
        contract_registry.specs.pop(str(configuration.public_id))

    # load into sys modules and register into contract registry
    contract = Contract.from_config(configuration)
    yield contract
    contract_registry.specs.pop(str(configuration.public_id))


def test_get_instance_no_address_ethereum(dummy_contract):
    """Tests get instance method with no address for ethereum."""
    ledger_api = ledger_apis_registry.make(ETHEREUM, address=ETHEREUM_DEFAULT_ADDRESS,)
    instance = dummy_contract.get_instance(ledger_api)
    assert type(instance) == web3._utils.datatypes.PropertyCheckingFactory


def test_get_deploy_transaction_ethereum(
    dummy_contract, ganache_addr, ganache_port, ganache
):
    """Tests the deploy transaction classmethod for ethereum."""
    ethereum_crypto = crypto_registry.make(ETHEREUM)
    ledger_api = ledger_apis_registry.make(
        ETHEREUM, address=make_uri(ganache_addr, ganache_port)
    )
    with patch(
        "web3.contract.ContractConstructor.buildTransaction",
        return_value={"data": "0xstub"},
    ):
        deploy_tx = dummy_contract.get_deploy_transaction(
            ledger_api, ethereum_crypto.address
        )
    assert deploy_tx is not None and len(deploy_tx) == 6
    assert all(
        key in ["from", "value", "gas", "gasPrice", "nonce", "data"]
        for key in deploy_tx.keys()
    )


def test_get_instance_no_address_cosmwasm(dummy_contract):
    """Tests get instance method with no address for fetchai."""
    ledger_api = ledger_apis_registry.make(FETCHAI, address=FETCHAI_DEFAULT_ADDRESS,)
    instance = dummy_contract.get_instance(ledger_api)
    assert instance is None


def test_get_deploy_transaction_cosmwasm(dummy_contract):
    """Tests the deploy transaction classmethod for fetchai."""
    fetchai_crypto = crypto_registry.make(FETCHAI)
    ledger_api = ledger_apis_registry.make(FETCHAI, address=FETCHAI_DEFAULT_ADDRESS,)
    deploy_tx = dummy_contract.get_deploy_transaction(
        ledger_api, fetchai_crypto.address
    )
    assert deploy_tx is not None and len(deploy_tx) == 6
    assert all(
        key in ["account_number", "chain_id", "fee", "memo", "msgs", "sequence"]
        for key in deploy_tx.keys()
    )


def test_scaffold():
    """Test the scaffold contract can be loaded/instantiated."""
    scaffold = MyScaffoldContract("config")
    kwargs = {"key": "value"}
    with pytest.raises(NotImplementedError):
        scaffold.get_raw_transaction("ledger_api", "contract_address", **kwargs)
    with pytest.raises(NotImplementedError):
        scaffold.get_raw_message("ledger_api", "contract_address", **kwargs)
    with pytest.raises(NotImplementedError):
        scaffold.get_state("ledger_api", "contract_address", **kwargs)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2020 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Module to declare constants."""
from pathlib import Path
from typing import List


FETCHAI = "fetchai"
DEFAULT_CONNECTION = "fetchai/stub:latest"
DEFAULT_PROTOCOL = "fetchai/default:latest"
SIGNING_PROTOCOL = "fetchai/signing:latest"
STATE_UPDATE_PROTOCOL = "fetchai/state_update:latest"
DEFAULT_SKILL = "fetchai/error:latest"
LEDGER_CONNECTION = "fetchai/ledger:latest"
DEFAULT_LEDGER = FETCHAI
PRIVATE_KEY_PATH_SCHEMA = "{}_private_key.txt"
DEFAULT_PRIVATE_KEY_FILE = PRIVATE_KEY_PATH_SCHEMA.format(DEFAULT_LEDGER)
DEFAULT_LICENSE = "Apache-2.0"
DISTRIBUTED_PACKAGES: List[str] = []
DEFAULT_SEARCH_SERVICE_ADDRESS = "fetchai/soef:any"
DEFAULT_INPUT_FILE_NAME = "./input_file"
DEFAULT_OUTPUT_FILE_NAME = "./output_file"
SCAFFOLD_PUBLIC_ID = "fetchai/scaffold:0.1.0"
PACKAGES = "packages"
DEFAULT_REGISTRY_NAME = PACKAGES
DEFAULT_REGISTRY_PATH = Path("./", DEFAULT_REGISTRY_NAME)
VENDOR = "vendor"
AGENT = "agent"
AGENTS = "agents"
CONNECTION = "connection"
CONNECTIONS = "connections"
CONTRACT = "contract"
CONTRACTS = "contracts"
PROTOCOL = "protocol"
PROTOCOLS = "protocols"
SKILL = "skill"
SKILLS = "skills"
DEFAULT_README_FILE = "README.md"
DEFAULT_VERSION = "0.1.0"
DEFAULT_AEA_CONFIG_FILE = "aea-config.yaml"
DEFAULT_SKILL_CONFIG_FILE = "skill.yaml"
DEFAULT_CONNECTION_CONFIG_FILE = "connection.yaml"
DEFAULT_CONTRACT_CONFIG_FILE = "contract.yaml"
DEFAULT_PROTOCOL_CONFIG_FILE = "protocol.yaml"
DEFAULT_LICENSE = "Apache-2.0"
PACKAGE_PUBLIC_ID_VAR_NAME = "PUBLIC_ID"
DEFAULT_FINGERPRINT_IGNORE_PATTERNS = [
    ".DS_Store",
    "*__pycache__/*",
    "*__pycache__",
    "*.pyc",
    DEFAULT_AEA_CONFIG_FILE,
    DEFAULT_PROTOCOL_CONFIG_FILE,
    DEFAULT_CONNECTION_CONFIG_FILE,
    DEFAULT_SKILL_CONFIG_FILE,
    DEFAULT_CONTRACT_CONFIG_FILE,
]
DEFAULT_PYPI_INDEX_URL = "https://pypi.org/simple"
DEFAULT_GIT_REF = "master"
IMPORT_TEMPLATE_1 = "from packages.{author}.{type}.{name}"
IMPORT_TEMPLATE_2 = "import packages.{author}.{type}.{name}"
DEFAULT_ENV_DOTFILE = ".env"



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the dialogue classes of the tac control contract skill."""

from pathlib import Path
from typing import cast

import pytest

from aea.exceptions import AEAEnforceError
from aea.helpers.transaction.base import RawTransaction, Terms
from aea.protocols.dialogue.base import DialogueLabel
from aea.test_tools.test_skill import BaseSkillTestCase, COUNTERPARTY_ADDRESS

from packages.fetchai.protocols.contract_api.custom_types import Kwargs
from packages.fetchai.protocols.contract_api.message import ContractApiMessage
from packages.fetchai.protocols.ledger_api.message import LedgerApiMessage
from packages.fetchai.protocols.signing.message import SigningMessage
from packages.fetchai.skills.tac_control_contract.dialogues import (
    ContractApiDialogue,
    ContractApiDialogues,
    LedgerApiDialogue,
    LedgerApiDialogues,
    SigningDialogue,
    SigningDialogues,
)

from tests.conftest import ROOT_DIR


class TestTacDialogues(BaseSkillTestCase):
    """Test dialogue classes of tac control contract."""

    path_to_skill = Path(
        ROOT_DIR, "packages", "fetchai", "skills", "tac_control_contract"
    )

    @classmethod
    def setup(cls):
        """Setup the test class."""
        super().setup()
        cls.contract_api_dialogues = cast(
            ContractApiDialogues, cls._skill.skill_context.contract_api_dialogues
        )
        cls.signing_dialogues = cast(
            SigningDialogues, cls._skill.skill_context.signing_dialogues
        )
        cls.ledger_api_dialogues = cast(
            LedgerApiDialogues, cls._skill.skill_context.ledger_api_dialogues
        )

    def test_contract_api_dialogue(self):
        """Test the ContractApiDialogue class."""
        contract_api_dialogue = ContractApiDialogue(
            DialogueLabel(
                ("", ""), COUNTERPARTY_ADDRESS, self.skill.skill_context.agent_address,
            ),
            self.skill.skill_context.agent_address,
            role=ContractApiDialogue.Role.AGENT,
        )

        # callable
        with pytest.raises(ValueError, match="Callable not set!"):
            assert contract_api_dialogue.callable

        callable = ContractApiDialogue.Callable.GET_DEPLOY_TRANSACTION
        contract_api_dialogue.callable = callable
        with pytest.raises(AEAEnforceError, match="Callable already set!"):
            contract_api_dialogue.callable = callable
        assert contract_api_dialogue.callable == callable

        # terms
        with pytest.raises(ValueError, match="Terms not set!"):
            assert contract_api_dialogue.terms
        terms = Terms(
            "some_ledger_id",
            self.skill.skill_context.agent_address,
            "counterprty",
            {"currency_id": 50},
            {"good_id": -10},
            "some_nonce",
        )
        contract_api_dialogue.terms = terms
        with pytest.raises(AEAEnforceError, match="Terms already set!"):
            contract_api_dialogue.terms = terms
        assert contract_api_dialogue.terms == terms

    def test_contract_api_dialogues(self):
        """Test the ContractApiDialogues class."""
        _, dialogue = self.contract_api_dialogues.create(
            counterparty=COUNTERPARTY_ADDRESS,
            performative=ContractApiMessage.Performative.GET_DEPLOY_TRANSACTION,
            ledger_id="some_ledger_id",
            contract_id="some_contract_id",
            callable="some_callable",
            kwargs=Kwargs({"some_key": "some_value"}),
        )
        assert dialogue.role == ContractApiDialogue.Role.AGENT
        assert dialogue.self_address == self.skill.skill_context.agent_address

    def test_ledger_api_dialogue(self):
        """Test the LedgerApiDialogue class."""
        ledger_api_dialogue = LedgerApiDialogue(
            DialogueLabel(
                ("", ""), COUNTERPARTY_ADDRESS, self.skill.skill_context.agent_address,
            ),
            self.skill.skill_context.agent_address,
            role=ContractApiDialogue.Role.AGENT,
        )

        # associated_signing_dialogue
        with pytest.raises(ValueError, match="Associated signing dialogue not set!"):
            assert ledger_api_dialogue.associated_signing_dialogue
        signing_dialogue = SigningDialogue(
            DialogueLabel(
                ("", ""), COUNTERPARTY_ADDRESS, self.skill.skill_context.agent_address,
            ),
            self.skill.skill_context.agent_address,
            role=SigningDialogue.Role.SKILL,
        )
        ledger_api_dialogue.associated_signing_dialogue = signing_dialogue
        with pytest.raises(
            AEAEnforceError, match="Associated signing dialogue already set!"
        ):
            ledger_api_dialogue.associated_signing_dialogue = signing_dialogue
        assert ledger_api_dialogue.associated_signing_dialogue == signing_dialogue

    def test_ledger_api_dialogues(self):
        """Test the LedgerApiDialogues class."""
        _, dialogue = self.ledger_api_dialogues.create(
            counterparty=COUNTERPARTY_ADDRESS,
            performative=LedgerApiMessage.Performative.GET_BALANCE,
            ledger_id="some_ledger_id",
            address="some_address",
        )
        assert dialogue.role == LedgerApiDialogue.Role.AGENT
        assert dialogue.self_address == self.skill.skill_context.agent_address

    def test_signing_dialogue(self):
        """Test the SigningDialogue class."""
        signing_dialogue = SigningDialogue(
            DialogueLabel(
                ("", ""), COUNTERPARTY_ADDRESS, self.skill.skill_context.agent_address,
            ),
            self.skill.skill_context.agent_address,
            role=ContractApiDialogue.Role.AGENT,
        )

        # associated_contract_api_dialogue
        with pytest.raises(
            ValueError, match="Associated contract api dialogue not set!"
        ):
            assert signing_dialogue.associated_contract_api_dialogue
        contract_api_dialogue = ContractApiDialogue(
            DialogueLabel(
                ("", ""), COUNTERPARTY_ADDRESS, self.skill.skill_context.agent_address,
            ),
            self.skill.skill_context.agent_address,
            role=ContractApiDialogue.Role.AGENT,
        )
        signing_dialogue.associated_contract_api_dialogue = contract_api_dialogue
        with pytest.raises(
            AEAEnforceError, match="Associated contract api dialogue already set!"
        ):
            signing_dialogue.associated_contract_api_dialogue = contract_api_dialogue
        assert (
            signing_dialogue.associated_contract_api_dialogue == contract_api_dialogue
        )

    def test_signing_dialogues(self):
        """Test the SigningDialogues class."""
        _, dialogue = self.signing_dialogues.create(
            counterparty=COUNTERPARTY_ADDRESS,
            performative=SigningMessage.Performative.SIGN_TRANSACTION,
            terms=Terms(
                "some_ledger_id",
                "some_sender_address",
                "some_counterparty_address",
                dict(),
                dict(),
                "some_nonce",
            ),
            raw_transaction=RawTransaction("some_ledger_id", "some_body"),
        )
        assert dialogue.role == SigningDialogue.Role.SKILL
        assert dialogue.self_address == str(self.skill.skill_context.skill_id)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the handler for the 'ml_train' skill."""

import pickle  # nosec
import uuid
from typing import Optional, cast

from aea.configurations.base import PublicId
from aea.helpers.transaction.base import Terms
from aea.protocols.base import Message
from aea.skills.base import Handler

from packages.fetchai.connections.ledger.base import (
    CONNECTION_ID as LEDGER_CONNECTION_PUBLIC_ID,
)
from packages.fetchai.protocols.default.message import DefaultMessage
from packages.fetchai.protocols.ledger_api.message import LedgerApiMessage
from packages.fetchai.protocols.ml_trade.message import MlTradeMessage
from packages.fetchai.protocols.oef_search.message import OefSearchMessage
from packages.fetchai.protocols.signing.message import SigningMessage
from packages.fetchai.skills.ml_train.dialogues import (
    DefaultDialogues,
    LedgerApiDialogue,
    LedgerApiDialogues,
    MlTradeDialogue,
    MlTradeDialogues,
    OefSearchDialogue,
    OefSearchDialogues,
    SigningDialogue,
    SigningDialogues,
)
from packages.fetchai.skills.ml_train.strategy import Strategy


DUMMY_DIGEST = "dummy_digest"
LEDGER_API_ADDRESS = str(LEDGER_CONNECTION_PUBLIC_ID)


class MlTradeHandler(Handler):
    """ML trade handler."""

    SUPPORTED_PROTOCOL = MlTradeMessage.protocol_id

    def setup(self) -> None:
        """
        Set up the handler.

        :return: None
        """
        pass

    def handle(self, message: Message) -> None:
        """
        Handle messages.

        :param message: the message
        :return: None
        """
        ml_trade_msg = cast(MlTradeMessage, message)

        # recover dialogue
        ml_trade_dialogues = cast(MlTradeDialogues, self.context.ml_trade_dialogues)
        ml_trade_dialogue = cast(
            MlTradeDialogue, ml_trade_dialogues.update(ml_trade_msg)
        )
        if ml_trade_dialogue is None:
            self._handle_unidentified_dialogue(ml_trade_msg)
            return

        # handle message
        if ml_trade_msg.performative == MlTradeMessage.Performative.TERMS:
            self._handle_terms(ml_trade_msg, ml_trade_dialogue)
        elif ml_trade_msg.performative == MlTradeMessage.Performative.DATA:
            self._handle_data(ml_trade_msg)
        else:
            self._handle_invalid(ml_trade_msg, ml_trade_dialogue)

    def teardown(self) -> None:
        """
        Teardown the handler.

        :return: None
        """
        pass

    def _handle_unidentified_dialogue(self, ml_trade_msg: MlTradeMessage) -> None:
        """
        Handle an unidentified dialogue.

        :param fipa_msg: the message
        """
        self.context.logger.info(
            "received invalid ml_trade message={}, unidentified dialogue.".format(
                ml_trade_msg
            )
        )
        default_dialogues = cast(DefaultDialogues, self.context.default_dialogues)
        default_msg, _ = default_dialogues.create(
            counterparty=ml_trade_msg.sender,
            performative=DefaultMessage.Performative.ERROR,
            error_code=DefaultMessage.ErrorCode.INVALID_DIALOGUE,
            error_msg="Invalid dialogue.",
            error_data={"ml_trade_message": ml_trade_msg.encode()},
        )
        self.context.outbox.put_message(message=default_msg)

    def _handle_terms(
        self, ml_trade_msg: MlTradeMessage, ml_trade_dialogue: MlTradeDialogue
    ) -> None:
        """
        Handle the terms of the request.

        :param ml_trade_msg: the ml trade message
        :param ml_trade_dialogue: the dialogue object
        :return: None
        """
        terms = ml_trade_msg.terms
        self.context.logger.info(
            "received terms message from {}: terms={}".format(
                ml_trade_msg.sender[-5:], terms.values
            )
        )

        strategy = cast(Strategy, self.context.strategy)
        acceptable = strategy.is_acceptable_terms(terms)
        affordable = strategy.is_affordable_terms(terms)
        if not acceptable and affordable:
            self.context.logger.info(
                "rejecting, terms are not acceptable and/or affordable"
            )
            return

        if strategy.is_ledger_tx:
            # construct a tx for settlement on the ledger
            ledger_api_dialogues = cast(
                LedgerApiDialogues, self.context.ledger_api_dialogues
            )
            ledger_api_msg, ledger_api_dialogue = ledger_api_dialogues.create(
                counterparty=LEDGER_API_ADDRESS,
                performative=LedgerApiMessage.Performative.GET_RAW_TRANSACTION,
                terms=Terms(
                    ledger_id=terms.values["ledger_id"],
                    sender_address=self.context.agent_addresses[
                        terms.values["ledger_id"]
                    ],
                    counterparty_address=terms.values["address"],
                    amount_by_currency_id={
                        terms.values["currency_id"]: -terms.values["price"]
                    },
                    is_sender_payable_tx_fee=True,
                    quantities_by_good_id={"ml_training_data": 1},
                    nonce=uuid.uuid4().hex,
                    fee_by_currency_id={terms.values["currency_id"]: 1},
                ),
            )
            ledger_api_dialogue = cast(LedgerApiDialogue, ledger_api_dialogue)
            ledger_api_dialogue.associated_ml_trade_dialogue = ml_trade_dialogue
            self.context.outbox.put_message(message=ledger_api_msg)
            self.context.logger.info(
                "requesting transfer transaction from ledger api..."
            )
        else:
            # accept directly with a dummy transaction digest, no settlement
            ml_accept = ml_trade_dialogue.reply(
                performative=MlTradeMessage.Performative.ACCEPT,
                target_message=ml_trade_msg,
                tx_digest=DUMMY_DIGEST,
                terms=terms,
            )
            self.context.outbox.put_message(message=ml_accept)
            self.context.logger.info("sending dummy transaction digest ...")

    def _handle_data(self, ml_trade_msg: MlTradeMessage) -> None:
        """
        Handle the data.

        :param ml_trade_msg: the ml trade message
        :return: None
        """
        terms = ml_trade_msg.terms
        payload = ml_trade_msg.payload
        data = pickle.loads(payload)  # nosec
        if data is None:
            self.context.logger.info(
                "received data message with no data from {}".format(
                    ml_trade_msg.sender[-5:]
                )
            )
        else:
            self.context.logger.info(
                "received data message from {}: data shape={}, terms={}".format(
                    ml_trade_msg.sender[-5:], data[0].shape, terms.values
                )
            )
            self.context.ml_model.update(data[0], data[1], 5)
            self.context.strategy.is_searching = True

    def _handle_invalid(
        self, ml_trade_msg: MlTradeMessage, ml_trade_dialogue: MlTradeDialogue
    ) -> None:
        """
        Handle a fipa message of invalid performative.

        :param ml_trade_msg: the message
        :param ml_trade_dialogue: the dialogue object
        :return: None
        """
        self.context.logger.warning(
            "cannot handle ml_trade message of performative={} in dialogue={}.".format(
                ml_trade_msg.performative, ml_trade_dialogue
            )
        )


class OEFSearchHandler(Handler):
    """The OEF search handler."""

    SUPPORTED_PROTOCOL = OefSearchMessage.protocol_id  # type: Optional[PublicId]

    def setup(self) -> None:
        """Call to setup the handler."""
        pass

    def handle(self, message: Message) -> None:
        """
        Implement the reaction to a message.

        :param message: the message
        :return: None
        """
        oef_search_msg = cast(OefSearchMessage, message)

        # recover dialogue
        oef_search_dialogues = cast(
            OefSearchDialogues, self.context.oef_search_dialogues
        )
        oef_search_dialogue = cast(
            Optional[OefSearchDialogue], oef_search_dialogues.update(oef_search_msg)
        )
        if oef_search_dialogue is None:
            self._handle_unidentified_dialogue(oef_search_msg)
            return

        # handle message
        if oef_search_msg.performative is OefSearchMessage.Performative.OEF_ERROR:
            self._handle_error(oef_search_msg, oef_search_dialogue)
        elif oef_search_msg.performative is OefSearchMessage.Performative.SEARCH_RESULT:
            self._handle_search(oef_search_msg, oef_search_dialogue)
        else:
            self._handle_invalid(oef_search_msg, oef_search_dialogue)

    def teardown(self) -> None:
        """
        Implement the handler teardown.

        :return: None
        """
        pass

    def _handle_unidentified_dialogue(self, oef_search_msg: OefSearchMessage) -> None:
        """
        Handle an unidentified dialogue.

        :param msg: the message
        """
        self.context.logger.info(
            "received invalid oef_search message={}, unidentified dialogue.".format(
                oef_search_msg
            )
        )

    def _handle_error(
        self, oef_search_msg: OefSearchMessage, oef_search_dialogue: OefSearchDialogue
    ) -> None:
        """
        Handle an oef search message.

        :param oef_search_msg: the oef search message
        :param oef_search_dialogue: the dialogue
        :return: None
        """
        self.context.logger.info(
            "received oef_search error message={} in dialogue={}.".format(
                oef_search_msg, oef_search_dialogue
            )
        )

    def _handle_search(
        self, oef_search_msg: OefSearchMessage, oef_search_dialogue: OefSearchDialogue
    ) -> None:
        """
        Handle the search response.

        :param agents: the agents returned by the search
        :return: None
        """
        if len(oef_search_msg.agents) == 0:
            self.context.logger.info(
                "found no agents in dialogue={}, continue searching.".format(
                    oef_search_dialogue
                )
            )
            return

        self.context.logger.info(
            "found agents={}, stopping search.".format(
                list(map(lambda x: x[-5:], oef_search_msg.agents)),
            )
        )
        strategy = cast(Strategy, self.context.strategy)
        strategy.is_searching = False
        query = strategy.get_service_query()
        ml_trade_dialogues = cast(MlTradeDialogues, self.context.ml_trade_dialogues)
        for idx, opponent_address in enumerate(oef_search_msg.agents):
            if idx >= strategy.max_negotiations:
                continue
            self.context.logger.info(
                "sending CFT to agent={}".format(opponent_address[-5:])
            )
            cft_msg, _ = ml_trade_dialogues.create(
                counterparty=opponent_address,
                performative=MlTradeMessage.Performative.CFP,
                query=query,
            )
            self.context.outbox.put_message(message=cft_msg)

    def _handle_invalid(
        self, oef_search_msg: OefSearchMessage, oef_search_dialogue: OefSearchDialogue
    ) -> None:
        """
        Handle an oef search message.

        :param oef_search_msg: the oef search message
        :param oef_search_dialogue: the dialogue
        :return: None
        """
        self.context.logger.warning(
            "cannot handle oef_search message of performative={} in dialogue={}.".format(
                oef_search_msg.performative, oef_search_dialogue,
            )
        )


class LedgerApiHandler(Handler):
    """Implement the ledger handler."""

    SUPPORTED_PROTOCOL = LedgerApiMessage.protocol_id  # type: Optional[PublicId]

    def setup(self) -> None:
        """Implement the setup for the handler."""
        pass

    def handle(self, message: Message) -> None:
        """
        Implement the reaction to a message.

        :param message: the message
        :return: None
        """
        ledger_api_msg = cast(LedgerApiMessage, message)

        # recover dialogue
        ledger_api_dialogues = cast(
            LedgerApiDialogues, self.context.ledger_api_dialogues
        )
        ledger_api_dialogue = cast(
            Optional[LedgerApiDialogue], ledger_api_dialogues.update(ledger_api_msg)
        )
        if ledger_api_dialogue is None:
            self._handle_unidentified_dialogue(ledger_api_msg)
            return

        # handle message
        if ledger_api_msg.performative is LedgerApiMessage.Performative.BALANCE:
            self._handle_balance(ledger_api_msg)
        elif (
            ledger_api_msg.performative is LedgerApiMessage.Performative.RAW_TRANSACTION
        ):
            self._handle_raw_transaction(ledger_api_msg, ledger_api_dialogue)
        elif (
            ledger_api_msg.performative
            == LedgerApiMessage.Performative.TRANSACTION_DIGEST
        ):
            self._handle_transaction_digest(ledger_api_msg, ledger_api_dialogue)
        elif ledger_api_msg.performative == LedgerApiMessage.Performative.ERROR:
            self._handle_error(ledger_api_msg, ledger_api_dialogue)
        else:
            self._handle_invalid(ledger_api_msg, ledger_api_dialogue)

    def teardown(self) -> None:
        """
        Implement the handler teardown.

        :return: None
        """
        pass

    def _handle_unidentified_dialogue(self, ledger_api_msg: LedgerApiMessage) -> None:
        """
        Handle an unidentified dialogue.

        :param msg: the message
        """
        self.context.logger.info(
            "received invalid ledger_api message={}, unidentified dialogue.".format(
                ledger_api_msg
            )
        )

    def _handle_balance(self, ledger_api_msg: LedgerApiMessage) -> None:
        """
        Handle a message of balance performative.

        :param ledger_api_message: the ledger api message
        :param ledger_api_dialogue: the ledger api dialogue
        """
        strategy = cast(Strategy, self.context.strategy)
        if ledger_api_msg.balance > 0:
            self.context.logger.info(
                "starting balance on {} ledger={}.".format(
                    strategy.ledger_id, ledger_api_msg.balance,
                )
            )
            strategy.is_searching = True
            strategy.balance = ledger_api_msg.balance
        else:
            self.context.logger.warning(
                "you have no starting balance on {} ledger!".format(strategy.ledger_id)
            )
            self.context.is_active = False

    def _handle_raw_transaction(
        self, ledger_api_msg: LedgerApiMessage, ledger_api_dialogue: LedgerApiDialogue
    ) -> None:
        """
        Handle a message of raw_transaction performative.

        :param ledger_api_message: the ledger api message
        :param ledger_api_dialogue: the ledger api dialogue
        """
        self.context.logger.info("received raw transaction={}".format(ledger_api_msg))
        signing_dialogues = cast(SigningDialogues, self.context.signing_dialogues)
        last_msg = cast(LedgerApiMessage, ledger_api_dialogue.last_outgoing_message)
        if last_msg is None:
            raise ValueError("Could not retrive last outgoing ledger_api_msg.")
        signing_msg, signing_dialogue = signing_dialogues.create(
            counterparty=self.context.decision_maker_address,
            performative=SigningMessage.Performative.SIGN_TRANSACTION,
            raw_transaction=ledger_api_msg.raw_transaction,
            terms=last_msg.terms,
        )
        signing_dialogue = cast(SigningDialogue, signing_dialogue)
        signing_dialogue.associated_ledger_api_dialogue = ledger_api_dialogue
        self.context.decision_maker_message_queue.put_nowait(signing_msg)
        self.context.logger.info(
            "proposing the transaction to the decision maker. Waiting for confirmation ..."
        )

    def _handle_transaction_digest(
        self, ledger_api_msg: LedgerApiMessage, ledger_api_dialogue: LedgerApiDialogue
    ) -> None:
        """
        Handle a message of transaction_digest performative.

        :param ledger_api_message: the ledger api message
        :param ledger_api_dialogue: the ledger api dialogue
        """
        ml_trade_dialogue = ledger_api_dialogue.associated_ml_trade_dialogue
        self.context.logger.info(
            "transaction was successfully submitted. Transaction digest={}".format(
                ledger_api_msg.transaction_digest
            )
        )
        ml_trade_msg = cast(
            Optional[MlTradeMessage], ml_trade_dialogue.last_incoming_message
        )
        if ml_trade_msg is None:
            raise ValueError("Could not retrieve ml_trade message")
        ml_accept = ml_trade_dialogue.reply(
            performative=MlTradeMessage.Performative.ACCEPT,
            target_message=ml_trade_msg,
            tx_digest=ledger_api_msg.transaction_digest.body,
            terms=ml_trade_msg.terms,
        )
        self.context.outbox.put_message(message=ml_accept)
        self.context.logger.info(
            "informing counterparty={} of transaction digest={}.".format(
                ml_trade_msg.sender[-5:], ledger_api_msg.transaction_digest,
            )
        )

    def _handle_error(
        self, ledger_api_msg: LedgerApiMessage, ledger_api_dialogue: LedgerApiDialogue
    ) -> None:
        """
        Handle a message of error performative.

        :param ledger_api_message: the ledger api message
        :param ledger_api_dialogue: the ledger api dialogue
        """
        self.context.logger.info(
            "received ledger_api error message={} in dialogue={}.".format(
                ledger_api_msg, ledger_api_dialogue
            )
        )

    def _handle_invalid(
        self, ledger_api_msg: LedgerApiMessage, ledger_api_dialogue: LedgerApiDialogue
    ) -> None:
        """
        Handle a message of invalid performative.

        :param ledger_api_message: the ledger api message
        :param ledger_api_dialogue: the ledger api dialogue
        """
        self.context.logger.warning(
            "cannot handle ledger_api message of performative={} in dialogue={}.".format(
                ledger_api_msg.performative, ledger_api_dialogue,
            )
        )


class SigningHandler(Handler):
    """Implement the transaction handler."""

    SUPPORTED_PROTOCOL = SigningMessage.protocol_id  # type: Optional[PublicId]

    def setup(self) -> None:
        """Implement the setup for the handler."""
        pass

    def handle(self, message: Message) -> None:
        """
        Implement the reaction to a message.

        :param message: the message
        :return: None
        """
        signing_msg = cast(SigningMessage, message)

        # recover dialogue
        signing_dialogues = cast(SigningDialogues, self.context.signing_dialogues)
        signing_dialogue = cast(
            Optional[SigningDialogue], signing_dialogues.update(signing_msg)
        )
        if signing_dialogue is None:
            self._handle_unidentified_dialogue(signing_msg)
            return

        # handle message
        if signing_msg.performative is SigningMessage.Performative.SIGNED_TRANSACTION:
            self._handle_signed_transaction(signing_msg, signing_dialogue)
        elif signing_msg.performative is SigningMessage.Performative.ERROR:
            self._handle_error(signing_msg, signing_dialogue)
        else:
            self._handle_invalid(signing_msg, signing_dialogue)

    def teardown(self) -> None:
        """
        Implement the handler teardown.

        :return: None
        """
        pass

    def _handle_unidentified_dialogue(self, signing_msg: SigningMessage) -> None:
        """
        Handle an unidentified dialogue.

        :param msg: the message
        """
        self.context.logger.info(
            "received invalid signing message={}, unidentified dialogue.".format(
                signing_msg
            )
        )

    def _handle_signed_transaction(
        self, signing_msg: SigningMessage, signing_dialogue: SigningDialogue
    ) -> None:
        """
        Handle an oef search message.

        :param signing_msg: the signing message
        :param signing_dialogue: the dialogue
        :return: None
        """
        self.context.logger.info("transaction signing was successful.")
        ledger_api_dialogue = signing_dialogue.associated_ledger_api_dialogue
        last_ledger_api_msg = cast(
            Optional[LedgerApiMessage], ledger_api_dialogue.last_incoming_message
        )
        if last_ledger_api_msg is None:
            raise ValueError("Could not retrieve last message in ledger api dialogue")
        ledger_api_msg = ledger_api_dialogue.reply(
            performative=LedgerApiMessage.Performative.SEND_SIGNED_TRANSACTION,
            target_message=last_ledger_api_msg,
            signed_transaction=signing_msg.signed_transaction,
        )
        self.context.outbox.put_message(message=ledger_api_msg)
        self.context.logger.info("sending transaction to ledger.")

    def _handle_error(
        self, signing_msg: SigningMessage, signing_dialogue: SigningDialogue
    ) -> None:
        """
        Handle an oef search message.

        :param signing_msg: the signing message
        :param signing_dialogue: the dialogue
        :return: None
        """
        self.context.logger.info(
            "transaction signing was not successful. Error_code={} in dialogue={}".format(
                signing_msg.error_code, signing_dialogue
            )
        )

    def _handle_invalid(
        self, signing_msg: SigningMessage, signing_dialogue: SigningDialogue
    ) -> None:
        """
        Handle an oef search message.

        :param signing_msg: the signing message
        :param signing_dialogue: the dialogue
        :return: None
        """
        self.context.logger.warning(
            "cannot handle signing message of performative={} in dialogue={}.".format(
                signing_msg.performative, signing_dialogue
            )
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Implementation of background profiling daemon."""

import asyncio
import datetime
import gc
import logging
import platform
import textwrap
import threading
import time
from collections import Counter
from concurrent.futures._base import CancelledError
from functools import wraps
from typing import Callable, Dict, List, Type

from aea.helpers.async_utils import Runnable


lock = threading.Lock()

_default_logger = logging.getLogger(__file__)

if platform.system() == "Windows":  # pragma: nocover
    import win32process  # type: ignore  # pylint: disable=import-error

    WIN32_PROCESS_TIMES_TICKS_PER_SECOND = 1e7

    def get_current_process_memory_usage() -> float:
        """Get current process memory usage in MB."""
        d = win32process.GetProcessMemoryInfo(win32process.GetCurrentProcess())  # type: ignore
        return 1.0 * d["WorkingSetSize"] / 1024 ** 2

    def get_current_process_cpu_time() -> float:
        """Get current process cpu time in seconds."""
        d = win32process.GetProcessTimes(win32process.GetCurrentProcess())  # type: ignore
        return d["UserTime"] / WIN32_PROCESS_TIMES_TICKS_PER_SECOND


else:
    import resource

    _MAC_MEM_STATS_MB = 1024 ** 2
    _LINUX_MEM_STATS_MB = 1024

    def get_current_process_memory_usage() -> float:
        """Get current process memory usage in MB."""
        if platform.system() == "Darwin":  # pragma: nocover
            divider = _MAC_MEM_STATS_MB
        else:
            divider = _LINUX_MEM_STATS_MB

        return 1.0 * resource.getrusage(resource.RUSAGE_SELF).ru_maxrss / divider

    def get_current_process_cpu_time() -> float:
        """Get current process cpu time in seconds."""
        return resource.getrusage(resource.RUSAGE_SELF).ru_utime


class Profiling(Runnable):
    """Profiling service."""

    def __init__(
        self,
        period: int = 0,
        objects_instances_to_count: List[Type] = None,
        objects_created_to_count: List[Type] = None,
        output_function: Callable[[str], None] = lambda x: print(x, flush=True),
    ) -> None:
        """
        Init profiler.

        :param period: delay between profiling output in seconds.
        :param output_function: function to display ouput, one str argument.
        """
        if period < 1:  # pragma: nocover
            raise ValueError("Period should be at least 1 second!")
        super().__init__(threaded=True)
        self._period = period
        self._start_ts = time.time()
        self._objects_instances_to_count = objects_instances_to_count or []
        self._objects_created_to_count = objects_created_to_count or []
        self._output_function = output_function
        self._counter: Dict[Type, int] = Counter()

    def set_counters(self):
        """Modify obj.__new__ to count objects created created."""
        for obj in self._objects_created_to_count:
            self._counter[obj] = 0

            def make_fn(obj) -> Callable:
                orig_new = obj.__new__
                # pylint: disable=protected-access  # type: ignore

                @wraps(orig_new)
                def new(*args, **kwargs):
                    self._counter[obj] += 1
                    if orig_new is object.__new__:
                        return orig_new(args[0])  # pragma: nocover
                    return orig_new(*args, **kwargs)  # pragma: nocover

                return new

            obj.__new__ = make_fn(obj)  # type: ignore

    async def run(self) -> None:
        """Run profiling."""
        try:
            self.set_counters()
            while True:
                await asyncio.sleep(self._period)
                self.output_profile_data()
        except CancelledError:
            pass
        except Exception:  # pragma: nocover
            _default_logger.exception("Exception in Profiling")
            raise

    def output_profile_data(self) -> None:
        """Render profiling data and call output_function."""
        data = self.get_profile_data()
        text = (
            textwrap.dedent(
                f"""
        Profiling details for current AEA process: {datetime.datetime.now()}
        =============================================
        Run time: {data["run_time"]:.6f} seconds
        Cpu time: {data["cpu_time"]:.6f} seconds,
        Cpu/Run time: {100*data["cpu_time"]/data["run_time"]:.6f}%
        Memory: {data["mem"]:.6f} MB
        Threads: {data["threads"]['amount']}  {data["threads"]['names']}
        Objects present:
        """
            )
            + "\n".join([f" * {i}:  {c}" for i, c in data["objects_present"].items()])
            + "\n"
            + """Objects created:\n"""
            + "\n".join(
                [f" * {i.__name__}:  {c}" for i, c in data["objects_created"].items()]
            )
            + "\n"
        )
        self._output_function(text)

    def get_profile_data(self) -> Dict:
        """Get profiling data dict."""
        return {
            "run_time": time.time() - self._start_ts,
            "cpu_time": get_current_process_cpu_time(),
            "mem": get_current_process_memory_usage(),
            "threads": {
                "amount": threading.active_count(),
                "names": [i.name for i in threading.enumerate()],
            },
            "objects_present": self.get_objects_instances(),
            "objects_created": self.get_objecst_created(),
        }

    def get_objects_instances(self) -> Dict:
        """Return dict with counted object instances present now."""
        result: Dict = Counter()

        lock.acquire()
        try:
            for obj_type in self._objects_instances_to_count:
                result[obj_type.__name__] += 0

            for obj in gc.get_objects():
                for obj_type in self._objects_instances_to_count:
                    if isinstance(obj, obj_type):
                        result[obj_type.__name__] += 1
        finally:
            lock.release()
        return result

    def get_objecst_created(self) -> Dict:
        """Return dict with counted object instances created."""
        return self._counter



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This test module contains tests for Libp2p tcp client connection."""

import os
import shutil
import tempfile

import pytest

from aea.mail.base import Envelope
from aea.multiplexer import Multiplexer

from packages.fetchai.connections.p2p_libp2p_client.connection import Uri
from packages.fetchai.protocols.default.message import DefaultMessage
from packages.fetchai.protocols.default.serialization import DefaultSerializer

from tests.conftest import (
    _make_libp2p_client_connection,
    _make_libp2p_connection,
    libp2p_log_on_failure,
    libp2p_log_on_failure_all,
)


DEFAULT_PORT = 10234
DEFAULT_DELEGATE_PORT = 11234
DEFAULT_HOST = "127.0.0.1"
DEFAULT_CLIENTS_PER_NODE = 4


@pytest.mark.asyncio
class TestLibp2pClientConnectionConnectDisconnect:
    """Test that connection is established and torn down correctly"""

    @classmethod
    def setup_class(cls):
        """Set the test up"""
        cls.cwd = os.getcwd()
        cls.t = tempfile.mkdtemp()
        os.chdir(cls.t)

        cls.connection_node = _make_libp2p_connection(delegate=True)
        cls.connection = _make_libp2p_client_connection()

    @pytest.mark.asyncio
    async def test_libp2pclientconnection_connect_disconnect(self):
        """Test connnect then disconnect."""
        assert self.connection.is_connected is False
        try:
            await self.connection_node.connect()
            await self.connection.connect()
            assert self.connection.is_connected is True

            await self.connection.disconnect()
            assert self.connection.is_connected is False
        except Exception:
            raise
        finally:
            await self.connection_node.disconnect()

    @classmethod
    def teardown_class(cls):
        """Tear down the test"""
        os.chdir(cls.cwd)
        try:
            shutil.rmtree(cls.t)
        except (OSError, IOError):
            pass


@libp2p_log_on_failure_all
class TestLibp2pClientConnectionEchoEnvelope:
    """Test that connection will route envelope to destination through the same libp2p node"""

    @classmethod
    @libp2p_log_on_failure
    def setup_class(cls):
        """Set the test up"""
        cls.cwd = os.getcwd()
        cls.t = tempfile.mkdtemp()
        os.chdir(cls.t)

        cls.log_files = []

        cls.connection_node = _make_libp2p_connection(DEFAULT_PORT + 1, delegate=True)
        cls.multiplexer_node = Multiplexer([cls.connection_node])
        cls.log_files.append(cls.connection_node.node.log_file)
        cls.multiplexer_node.connect()

        try:
            cls.connection_client_1 = _make_libp2p_client_connection()
            cls.multiplexer_client_1 = Multiplexer([cls.connection_client_1])
            cls.multiplexer_client_1.connect()

            cls.connection_client_2 = _make_libp2p_client_connection()
            cls.multiplexer_client_2 = Multiplexer([cls.connection_client_2])
            cls.multiplexer_client_2.connect()
        except Exception:
            cls.multiplexer_node.disconnect()
            raise

    def test_connection_is_established(self):
        """Test connection is established."""
        assert self.connection_client_1.is_connected is True
        assert self.connection_client_2.is_connected is True

    def test_envelope_routed(self):
        """Test the envelope is routed."""
        addr_1 = self.connection_client_1.address
        addr_2 = self.connection_client_2.address

        msg = DefaultMessage(
            dialogue_reference=("", ""),
            message_id=1,
            target=0,
            performative=DefaultMessage.Performative.BYTES,
            content=b"hello",
        )
        envelope = Envelope(
            to=addr_2,
            sender=addr_1,
            protocol_id=DefaultMessage.protocol_id,
            message=DefaultSerializer().encode(msg),
        )

        self.multiplexer_client_1.put(envelope)
        delivered_envelope = self.multiplexer_client_2.get(block=True, timeout=20)

        assert delivered_envelope is not None
        assert delivered_envelope.to == envelope.to
        assert delivered_envelope.sender == envelope.sender
        assert delivered_envelope.protocol_id == envelope.protocol_id
        assert delivered_envelope.message == envelope.message

    def test_envelope_echoed_back(self):
        """Test the envelope is echoed back."""
        addr_1 = self.connection_client_1.address
        addr_2 = self.connection_client_2.address

        msg = DefaultMessage(
            dialogue_reference=("", ""),
            message_id=1,
            target=0,
            performative=DefaultMessage.Performative.BYTES,
            content=b"hello",
        )
        original_envelope = Envelope(
            to=addr_2,
            sender=addr_1,
            protocol_id=DefaultMessage.protocol_id,
            message=DefaultSerializer().encode(msg),
        )

        self.multiplexer_client_1.put(original_envelope)
        delivered_envelope = self.multiplexer_client_2.get(block=True, timeout=10)
        assert delivered_envelope is not None

        delivered_envelope.to = addr_1
        delivered_envelope.sender = addr_2

        self.multiplexer_client_2.put(delivered_envelope)
        echoed_envelope = self.multiplexer_client_1.get(block=True, timeout=5)

        assert echoed_envelope is not None
        assert echoed_envelope.to == original_envelope.sender
        assert delivered_envelope.sender == original_envelope.to
        assert delivered_envelope.protocol_id == original_envelope.protocol_id
        assert delivered_envelope.message == original_envelope.message

    def test_envelope_echoed_back_node_agent(self):
        """Test the envelope is echoed back."""
        addr_1 = self.connection_client_1.address
        addr_n = self.connection_node.address

        msg = DefaultMessage(
            dialogue_reference=("", ""),
            message_id=1,
            target=0,
            performative=DefaultMessage.Performative.BYTES,
            content=b"hello",
        )
        original_envelope = Envelope(
            to=addr_n,
            sender=addr_1,
            protocol_id=DefaultMessage.protocol_id,
            message=DefaultSerializer().encode(msg),
        )

        self.multiplexer_client_1.put(original_envelope)
        delivered_envelope = self.multiplexer_node.get(block=True, timeout=10)
        assert delivered_envelope is not None

        delivered_envelope.to = addr_1
        delivered_envelope.sender = addr_n

        self.multiplexer_node.put(delivered_envelope)
        echoed_envelope = self.multiplexer_client_1.get(block=True, timeout=5)

        assert echoed_envelope is not None
        assert echoed_envelope.to == original_envelope.sender
        assert delivered_envelope.sender == original_envelope.to
        assert delivered_envelope.protocol_id == original_envelope.protocol_id
        assert delivered_envelope.message == original_envelope.message

    @classmethod
    def teardown_class(cls):
        """Tear down the test"""
        cls.multiplexer_client_1.disconnect()
        cls.multiplexer_client_2.disconnect()
        cls.multiplexer_node.disconnect()

        os.chdir(cls.cwd)
        try:
            shutil.rmtree(cls.t)
        except (OSError, IOError):
            pass


@libp2p_log_on_failure_all
class TestLibp2pClientConnectionEchoEnvelopeTwoDHTNode:
    """Test that connection will route envelope to destination connected to different node"""

    @classmethod
    @libp2p_log_on_failure
    def setup_class(cls):
        """Set the test up"""
        cls.cwd = os.getcwd()
        cls.t = tempfile.mkdtemp()
        os.chdir(cls.t)

        cls.log_files = []
        cls.mutliplexers = []

        cls.connection_node_1 = _make_libp2p_connection(
            port=DEFAULT_PORT + 1,
            delegate_port=DEFAULT_DELEGATE_PORT + 1,
            delegate=True,
        )
        cls.multiplexer_node_1 = Multiplexer([cls.connection_node_1])
        cls.log_files.append(cls.connection_node_1.node.log_file)
        cls.multiplexer_node_1.connect()
        cls.mutliplexers.append(cls.multiplexer_node_1)

        genesis_peer = cls.connection_node_1.node.multiaddrs[0]

        try:
            cls.connection_node_2 = _make_libp2p_connection(
                port=DEFAULT_PORT + 2,
                delegate_port=DEFAULT_DELEGATE_PORT + 2,
                entry_peers=[genesis_peer],
                delegate=True,
            )
            cls.multiplexer_node_2 = Multiplexer([cls.connection_node_2])
            cls.log_files.append(cls.connection_node_2.node.log_file)
            cls.multiplexer_node_2.connect()
            cls.mutliplexers.append(cls.multiplexer_node_2)

            cls.connection_client_1 = _make_libp2p_client_connection(
                DEFAULT_DELEGATE_PORT + 1
            )
            cls.multiplexer_client_1 = Multiplexer([cls.connection_client_1])
            cls.multiplexer_client_1.connect()
            cls.mutliplexers.append(cls.multiplexer_client_1)

            cls.connection_client_2 = _make_libp2p_client_connection(
                DEFAULT_DELEGATE_PORT + 2
            )
            cls.multiplexer_client_2 = Multiplexer([cls.connection_client_2])
            cls.multiplexer_client_2.connect()
            cls.mutliplexers.append(cls.multiplexer_client_2)
        except Exception:
            cls.teardown_class()
            raise

    def test_connection_is_established(self):
        """Test the connection is established."""
        assert self.connection_node_1.is_connected is True
        assert self.connection_node_2.is_connected is True
        assert self.connection_client_1.is_connected is True
        assert self.connection_client_2.is_connected is True

    def test_envelope_routed(self):
        """Test the envelope is routed."""
        addr_1 = self.connection_client_1.address
        addr_2 = self.connection_client_2.address

        msg = DefaultMessage(
            dialogue_reference=("", ""),
            message_id=1,
            target=0,
            performative=DefaultMessage.Performative.BYTES,
            content=b"hello",
        )
        envelope = Envelope(
            to=addr_2,
            sender=addr_1,
            protocol_id=DefaultMessage.protocol_id,
            message=DefaultSerializer().encode(msg),
        )

        self.multiplexer_client_1.put(envelope)
        delivered_envelope = self.multiplexer_client_2.get(block=True, timeout=20)

        assert delivered_envelope is not None
        assert delivered_envelope.to == envelope.to
        assert delivered_envelope.sender == envelope.sender
        assert delivered_envelope.protocol_id == envelope.protocol_id
        assert delivered_envelope.message == envelope.message

    def test_envelope_echoed_back(self):
        """Test the envelope is echoed back."""
        addr_1 = self.connection_client_1.address
        addr_2 = self.connection_client_2.address

        msg = DefaultMessage(
            dialogue_reference=("", ""),
            message_id=1,
            target=0,
            performative=DefaultMessage.Performative.BYTES,
            content=b"hello",
        )
        original_envelope = Envelope(
            to=addr_2,
            sender=addr_1,
            protocol_id=DefaultMessage.protocol_id,
            message=DefaultSerializer().encode(msg),
        )

        self.multiplexer_client_1.put(original_envelope)
        delivered_envelope = self.multiplexer_client_2.get(block=True, timeout=10)
        assert delivered_envelope is not None

        delivered_envelope.to = addr_1
        delivered_envelope.sender = addr_2

        self.multiplexer_client_2.put(delivered_envelope)
        echoed_envelope = self.multiplexer_client_1.get(block=True, timeout=5)

        assert echoed_envelope is not None
        assert echoed_envelope.to == original_envelope.sender
        assert delivered_envelope.sender == original_envelope.to
        assert delivered_envelope.protocol_id == original_envelope.protocol_id
        assert delivered_envelope.message == original_envelope.message

    def test_envelope_echoed_back_node_agent(self):
        """Test the envelope is echoed back node agent."""
        addr_1 = self.connection_client_1.address
        addr_n = self.connection_node_2.address

        msg = DefaultMessage(
            dialogue_reference=("", ""),
            message_id=1,
            target=0,
            performative=DefaultMessage.Performative.BYTES,
            content=b"hello",
        )
        original_envelope = Envelope(
            to=addr_n,
            sender=addr_1,
            protocol_id=DefaultMessage.protocol_id,
            message=DefaultSerializer().encode(msg),
        )

        self.multiplexer_client_1.put(original_envelope)
        delivered_envelope = self.multiplexer_node_2.get(block=True, timeout=10)
        assert delivered_envelope is not None

        delivered_envelope.to = addr_1
        delivered_envelope.sender = addr_n

        self.multiplexer_node_2.put(delivered_envelope)
        echoed_envelope = self.multiplexer_client_1.get(block=True, timeout=5)

        assert echoed_envelope is not None
        assert echoed_envelope.to == original_envelope.sender
        assert delivered_envelope.sender == original_envelope.to
        assert delivered_envelope.protocol_id == original_envelope.protocol_id
        assert delivered_envelope.message == original_envelope.message

    @classmethod
    def teardown_class(cls):
        """Tear down the test"""
        for mux in reversed(cls.mutliplexers):
            mux.disconnect()
        os.chdir(cls.cwd)
        try:
            shutil.rmtree(cls.t)
        except (OSError, IOError):
            pass


@libp2p_log_on_failure_all
class TestLibp2pClientConnectionRouting:
    """Test that libp2p DHT network will reliably route envelopes from clients connected to different nodes"""

    @classmethod
    @libp2p_log_on_failure
    def setup_class(cls):
        """Set the test up"""
        cls.cwd = os.getcwd()
        cls.t = tempfile.mkdtemp()
        os.chdir(cls.t)

        cls.log_files = []
        cls.multiplexers = []

        try:
            cls.connection_node_1 = _make_libp2p_connection(
                port=DEFAULT_PORT + 1,
                delegate_port=DEFAULT_DELEGATE_PORT + 1,
                delegate=True,
            )
            cls.multiplexer_node_1 = Multiplexer([cls.connection_node_1])
            cls.log_files.append(cls.connection_node_1.node.log_file)
            cls.multiplexer_node_1.connect()
            cls.multiplexers.append(cls.multiplexer_node_1)

            entry_peer = cls.connection_node_1.node.multiaddrs[0]

            cls.connection_node_2 = _make_libp2p_connection(
                port=DEFAULT_PORT + 2,
                delegate_port=DEFAULT_DELEGATE_PORT + 2,
                entry_peers=[entry_peer],
                delegate=True,
            )
            cls.multiplexer_node_2 = Multiplexer([cls.connection_node_2])
            cls.log_files.append(cls.connection_node_2.node.log_file)
            cls.multiplexer_node_2.connect()
            cls.multiplexers.append(cls.multiplexer_node_2)

            cls.connections = [cls.connection_node_1, cls.connection_node_2]
            cls.addresses = [
                cls.connection_node_1.address,
                cls.connection_node_2.address,
            ]

            for _ in range(DEFAULT_CLIENTS_PER_NODE):
                for port in [DEFAULT_DELEGATE_PORT + 1, DEFAULT_DELEGATE_PORT + 2]:
                    conn = _make_libp2p_client_connection(port)
                    mux = Multiplexer([conn])

                    cls.connections.append(conn)
                    cls.addresses.append(conn.address)

                    mux.connect()
                    cls.multiplexers.append(mux)
        except Exception:
            cls.teardown_class()
            raise

    def test_connection_is_established(self):
        """Test connection is established."""
        for conn in self.connections:
            assert conn.is_connected is True

    def test_star_routing_connectivity(self):
        """Test routing with star connectivity."""
        msg = DefaultMessage(
            dialogue_reference=("", ""),
            message_id=1,
            target=0,
            performative=DefaultMessage.Performative.BYTES,
            content=b"hello",
        )

        for source in range(len(self.multiplexers)):
            for destination in range(len(self.multiplexers)):
                if destination == source:
                    continue
                envelope = Envelope(
                    to=self.addresses[destination],
                    sender=self.addresses[source],
                    protocol_id=DefaultMessage.protocol_id,
                    message=DefaultSerializer().encode(msg),
                )

                self.multiplexers[source].put(envelope)
                delivered_envelope = self.multiplexers[destination].get(
                    block=True, timeout=10
                )

                assert delivered_envelope is not None
                assert delivered_envelope.to == envelope.to
                assert delivered_envelope.sender == envelope.sender
                assert delivered_envelope.protocol_id == envelope.protocol_id
                assert delivered_envelope.message == envelope.message

    @classmethod
    def teardown_class(cls):
        """Tear down the test"""
        for mux in reversed(cls.multiplexers):
            mux.disconnect()
        os.chdir(cls.cwd)
        try:
            shutil.rmtree(cls.t)
        except (OSError, IOError):
            pass


def test_libp2pclientconnection_uri():
    """Test the uri."""
    uri = Uri(host="127.0.0.1")
    uri = Uri(host="127.0.0.1", port=10000)
    assert uri.host == "127.0.0.1" and uri.port == 10000



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Deploy an ACN libp2p node to a kubernetes cluster"""

import argparse
import base64
import os
import subprocess  # nosec
import sys
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Type


SCRIPT_DIR = os.path.abspath(os.path.dirname(os.path.realpath(__file__)))

sys.path.append(SCRIPT_DIR)
from run_acn_node_standalone import (  # noqa # pylint: disable=wrong-import-position # isort:skip
    AcnNodeConfig,
)

# docker defaults
DOCKER_FETCHAI_DEFAULT_FILE = os.path.join(SCRIPT_DIR, "Dockerfile")
DOCKER_FETCHAI_DEFAULT_FILE_DEV = os.path.join(SCRIPT_DIR, "Dockerfile.dev")
DOCKER_FETCHAI_DEFAULT_CTX = SCRIPT_DIR
DOCKER_FETCHAI_DEFAULT_CTX_DEV = os.path.join(SCRIPT_DIR, "../../")
DOCKER_FETCHAI_DEFAULT_IMG = "acn_node"
DOCKER_FETCHAI_DEFAULT_REGISTRY = "gcr.io/fetch-ai-sandbox"

# k8s defaults
K8S_FETCHAI_DEFAULT_PUBLIC_HOST = "agents-p2p-dht.sandbox.fetch-ai.com"
K8S_FETCHAI_DEFAULT_PUBLIC_TEMPLATE_DIR = os.path.join(SCRIPT_DIR, "k8s")
K8S_FETCHAI_DEFAULT_NAMESPACE = "agents-p2p-dht"


def _execute_cmd(cmd: List[str]) -> Tuple[str, bool]:
    """
    Run command as subprocess and wait for its termination

    :param cmd: command with arguments to execute
    :return: output of the command execution and execution success
    """

    print("-> Running: {}".format(cmd))
    proc = subprocess.Popen(cmd, shell=False, stdout=subprocess.PIPE)  # nosec
    out, _ = proc.communicate()
    success = False
    try:
        if proc.wait() == 0:
            success = True
    except (
        subprocess.CalledProcessError,  # pylint: disable=broad-except
        Exception,
    ) as e:
        print("_excute_cmd caught exception: {}".format(str(e)))
    print("-| Output :\n{}".format(out.decode("ascii")))
    return out.decode("ascii"), success


class DockerDeployment:
    """Build and Publish a Dockerfile"""

    def __init__(
        self, dockerfile: str, context: str, image: str, tag: str, registry: str
    ):
        """
        Initialize a DockerDeployment object

        :param dockerfile: path to the Dockerfile
        :param context: path to docker image context
        :param image: built image name
        :param tag: built image tag
        :param registry: registry to publish the image to
        """

        self.dockerfile = dockerfile
        self.context = context
        self.image = image
        self.tag = tag
        self.registry = registry

    def build_and_publish(self) -> bool:
        """
        Build the image and publich it to registry

        :return: success of the operation
        """

        cmds: List[List[str]] = list()

        cmds.append(
            [
                "docker",
                "build",
                "-t",
                "{}:{}".format(self.image, self.tag),
                "-f",
                self.dockerfile,
                self.context,
            ]
        )
        cmds.append(
            [
                "docker",
                "tag",
                "{}:{}".format(self.image, self.tag),
                "{}/{}:{}".format(self.registry, self.image, self.tag),
            ]
        )
        cmds.append(
            ["docker", "push", "{}/{}:{}".format(self.registry, self.image, self.tag)]
        )

        for cmd in cmds:
            _, ok = _execute_cmd(cmd)
            if not ok:
                return False

        return True


class K8sPodDeployment:
    """An application-agnostic deployment object"""

    def __init__(
        self,
        deployments_files: List[Path],
        docker_deployment: Optional[DockerDeployment],
    ):
        """
        Initialize a K8sPodDeployment object

        :param deployment_files: list of kubernetes yaml files to deploy
        :param docker_deployment: optional DockerDeployment to build and publish
        """

        self.deployment_files = deployments_files
        self.docker_deployment = docker_deployment

    def deploy(self) -> bool:
        """
        Deploy to k8s cluster

        :return: success of the operation
        """

        ok = True
        if self.docker_deployment:
            ok = self.docker_deployment.build_and_publish()
            if not ok:
                return ok
        for yaml in self.deployment_files:
            cmd = ["kubectl", "apply", "-f", str(yaml)]
            _, ok = _execute_cmd(cmd)
            if not ok:
                break
        return ok

    def delete(self) -> bool:
        """
        Delete deployment from k8s cluster

        :return: success of the operation
        """
        ok = True
        for yaml in self.deployment_files:
            cmd = ["kubectl", "delete", "-f", str(yaml)]
            _, ok = _execute_cmd(cmd)
            if not ok:
                break
        return ok


class AcnK8sPodConfig:
    """Store, parse, and generate kubernetes deployment for an ACN node"""

    K8S_DEPLOYMENT_NAME = "ph-deployment-name-here"
    K8S_NAMESPACE = "ph-deployment-namespace-here"
    K8S_PUBLIC_DNS = "ph-deployment-dns-here"

    DOCKER_IMAGE_REMOTE_WITH_TAG = "ph-gcr-image-with-tag-here"

    NODE_PORT = "ph-node-port-number-here"
    NODE_PORT_DELEGATE = "ph-node-delegate-port-number-here"
    NODE_PORT_MONITORING = "ph-node-monitoring-port-number-here"
    NODE_URI_EXTERNAL = "ph-node-external-uri-here"
    NODE_URI = "ph-node-local-uri-here"
    NODE_URI_DELEGATE = "ph-node-delegate-uri-here"
    NODE_URI_MONITORING = "ph-node-monitoring-uri-here"
    NODE_ENTRY_PEERS = "ph-node-entry-peers-list-here"
    NODE_KEY_NAME = "ph-node-priv-key-name-here"
    NODE_KEY_ENCODED = "ph-base64-encoded-private-key-here"
    NODE_LAST_ENTRY_PEER_HOST = "ph-latest-entry-peer-host-here"
    NODE_LAST_ENTRY_PEER_PORT = "ph-latest-entry-peer-port-here"
    NODE_LOG_FILE = "ph-node-log-file-path-here"
    # class defaults
    Defaults: Dict[str, str] = {
        K8S_DEPLOYMENT_NAME: "acn-node",
        NODE_LOG_FILE: "/acn_data/libp2p_node",
    }

    def __init__(
        self,
        acn_key_file: str,
        acn_port: int,
        acn_delegate_port: int,
        acn_monitoring_port: int,
        acn_entry_peers: Optional[List[str]],
        docker_file: str,
        docker_context: str,
        docker_image: str,
        docker_registry: str,
        k8s_public_hostname: str,
        k8s_namespace: str,
        k8s_template_files_dir: str,
        enable_checks: bool = True,
    ):
        """
        Initialize a AcnK8sPodConfig, populate the config dictionary

        :param acn_key_file: path to the acn node private key
        :param acn_port: acn node port number
        :param acn_delegate_port: acn node delegate service port number
        :param acn_monitoring_port: acn node monitoring service port number
        :param acn_entry_peers: optional list of acn node entry peers multiaddresses
        :param docker_file: path to Dockerfile
        :param docker_context: path to Dockerfile context
        :param docker_image: docker image name
        :param docker_registry: url of remote docker registry to push image to
        :param k8s_public_hostname: public dns for acn node's external uri
        :param k8s_namespace: k8s namespace to deploy node to
        :param k8s_template_files_dir: path to directory containing k8s yaml deployment templates
        :param enable_checks: enable configuration checks
        """

        config: Dict[str, str] = dict()
        cls: Type[AcnK8sPodConfig] = AcnK8sPodConfig

        # acn node configuration
        config[cls.NODE_KEY_NAME] = "node-priv-key-{}".format(acn_port)
        config[cls.NODE_PORT] = str(acn_port)
        config[cls.NODE_PORT_DELEGATE] = str(acn_delegate_port)
        config[cls.NODE_PORT_MONITORING] = str(acn_monitoring_port)
        config[cls.NODE_ENTRY_PEERS] = (
            ",".join(acn_entry_peers) if acn_entry_peers is not None else ""
        )
        config[cls.NODE_ENTRY_PEERS] = '"{}"'.format(config[cls.NODE_ENTRY_PEERS])
        peer_host, peer_port = cls._uri_from_multiaddr(
            acn_entry_peers[-1]
            if acn_entry_peers is not None and len(acn_entry_peers) > 0
            else ""
        )
        config[cls.NODE_LAST_ENTRY_PEER_HOST] = '"{}"'.format(peer_host)
        config[cls.NODE_LAST_ENTRY_PEER_PORT] = '"{}"'.format(peer_port)

        config[cls.NODE_URI] = "0.0.0.0:{}".format(acn_port)
        config[cls.NODE_URI_DELEGATE] = "0.0.0.0:{}".format(acn_delegate_port)
        config[cls.NODE_URI_MONITORING] = "0.0.0.0:{}".format(acn_monitoring_port)
        config[cls.NODE_URI_EXTERNAL] = "{}:{}".format(k8s_public_hostname, acn_port)
        config[cls.NODE_LOG_FILE] = '"{}_{}.log"'.format(
            cls.Defaults[cls.NODE_LOG_FILE], str(acn_port)
        )

        with open(acn_key_file, "r") as f:
            key = f.read().strip()
            config[cls.NODE_KEY_ENCODED] = base64.b64encode(key.encode("ascii")).decode(
                "ascii"
            )

        # k8s configuration
        config[cls.K8S_DEPLOYMENT_NAME] = "{}-{}".format(
            cls.Defaults[cls.K8S_DEPLOYMENT_NAME], str(acn_port)
        )
        config[cls.K8S_NAMESPACE] = k8s_namespace
        config[cls.K8S_PUBLIC_DNS] = k8s_public_hostname

        files: List[Path] = []
        for path in [
            Path(os.path.join(k8s_template_files_dir, p))
            for p in os.listdir(k8s_template_files_dir)
        ]:
            if path.is_file() and path.suffix == ".yaml":
                files.append(path)
        assert (
            len(files) > 0
        ), f"Couldn't find any template deployment file at {k8s_template_files_dir}"

        # docker configuration
        cmd = ["git", "describe", "--no-match", "--always", "--dirty"]
        docker_tag, ok = _execute_cmd(cmd)
        if not ok:
            docker_tag = datetime.now().strftime("%d-%m-%Y_%H-%M-%S")
        else:
            docker_tag = docker_tag.strip()
        config[cls.DOCKER_IMAGE_REMOTE_WITH_TAG] = "{}/{}:{}".format(
            docker_registry, docker_image, docker_tag.strip()
        )

        self.config = config
        self.template_files = files
        self.docker_deployment = DockerDeployment(
            docker_file, docker_context, docker_image, docker_tag, docker_registry
        )

        if enable_checks:
            cls.check_config(self.config)

    @staticmethod
    def _uri_from_multiaddr(maddr: str) -> Tuple[str, str]:
        if maddr != "":
            parts = maddr.split("/")
            if len(parts) == 7:
                return parts[2], parts[4]
        return "", ""

    @staticmethod
    def check_config(config: Dict[str, str]) -> None:
        """
        Check an AcnK8sPodConfig deployment for correct configuration

        :param config: dictionary of configuration to check
        """
        AcnNodeConfig(
            base64.b64decode(
                config[AcnK8sPodConfig.NODE_KEY_ENCODED].encode("ascii")
            ).decode("ascii"),
            config[AcnK8sPodConfig.NODE_URI],
            config[AcnK8sPodConfig.NODE_URI_EXTERNAL],
            config[AcnK8sPodConfig.NODE_URI_DELEGATE],
            config[AcnK8sPodConfig.NODE_URI_MONITORING],
            config[AcnK8sPodConfig.NODE_ENTRY_PEERS].strip('"').split(","),
            "",
            True,
        )

    def generate_deployment(self) -> K8sPodDeployment:
        """
        Generate deployment for the current configuration

        :return: deployment object
        """

        deployment_file = ".acn_deployment.yaml"
        out = open(deployment_file, "w")

        for path in self.template_files:
            with open(path, "r") as f:
                content = f.read()

            for placeholder, value in self.config.items():
                content = content.replace(placeholder, value)

            out.write(content)
            out.write("\n---\n")

        out.close()

        return K8sPodDeployment([Path(deployment_file)], self.docker_deployment)


def parse_commandline():
    """Parse script cl arguments"""

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--from-file",
        action="store",
        type=str,
        dest="from_file",
        required=False,
        help="Use previously generated deployment",
    )
    parser.add_argument(
        "--delete",
        action="store_true",
        dest="delete_deployment",
        required=False,
        help="Delete an already deployed node with the same configuration",
    )
    parser.add_argument(
        "--generate-only",
        action="store_true",
        dest="generate_only",
        required=False,
        help="Don't deploy the generated yaml file",
    )
    parser.add_argument(
        "--acn-key-file",
        action="store",
        type=str,
        dest="key",
        required=False,
        help="acn node's private key file",
    )
    parser.add_argument(
        "--acn-port",
        action="store",
        type=int,
        dest="port",
        required=False,
        help="acn node's port number (both local and external)",
    )
    parser.add_argument(
        "--acn-port-delegate",
        action="store",
        type=int,
        dest="delegate_port",
        required=False,
        help="acn node's delegate service port number (both local and external)",
    )
    parser.add_argument(
        "--acn-port-monitoring",
        action="store",
        type=int,
        dest="monitoring_port",
        required=False,
        help="acn node's monitoring service port number (local only)",
    )
    parser.add_argument(
        "--acn-entry-peers-maddrs",
        action="store",
        nargs="*",
        dest="entry_peers_maddrs",
        help="acn node's entry peers in libp2p multiaddress format",
    )
    parser.add_argument(
        "--k8s-fetchai-defaults",
        action="store_true",
        dest="k8s_fetchai_defaults",
        required=False,
        help="Use FetchAI defaults for k8s configuration",
    )
    parser.add_argument(
        "--k8s-public-hostname",
        action="store",
        type=str,
        dest="k8s_public_hostname",
        required=False,
        help="K8s public hostname to use for acn node's external uri",
    )
    parser.add_argument(
        "--k8s-namespace",
        action="store",
        type=str,
        dest="k8s_namespace",
        required=False,
        help="K8s deployment namespace",
    )
    parser.add_argument(
        "--k8s-template-files-dir",
        action="store",
        type=str,
        dest="k8s_template_files_dir",
        required=False,
        help="Directory containing k8s template yaml deployment files",
    )

    parser.add_argument(
        "--docker-fetchai-defaults",
        action="store_true",
        dest="docker_fetchai_defaults",
        required=False,
        help="Use FetchAI defaults for docker configuration",
    )
    parser.add_argument(
        "--docker-fetchai-defaults-dev",
        action="store_true",
        dest="docker_fetchai_defaults_dev",
        required=False,
        help="Use FetchAI Dev defaults for docker configuration",
    )
    parser.add_argument(
        "--docker-file",
        action="store",
        type=str,
        dest="docker_file",
        required=False,
        help="Path to Dockerfile to build",
    )
    parser.add_argument(
        "--docker-ctx",
        action="store",
        type=str,
        dest="docker_ctx",
        required=False,
        help="Path to Dockerfile context",
    )
    parser.add_argument(
        "--docker-image",
        action="store",
        type=str,
        dest="docker_image",
        required=False,
        help="Docker image name",
    )
    parser.add_argument(
        "--docker-registry",
        action="store",
        type=str,
        dest="docker_registry",
        required=False,
        help="Docker remote registry",
    )

    args = parser.parse_args()

    # checks
    if args.from_file is None and (
        args.key is None
        or args.port is None
        or args.delegate_port is None
        or args.monitoring_port is None
    ):
        parser.error(
            "--acn-key-file, --acn-port, --acn-port-delegate, --acn-port-monitoring are required when --from-file is not used"
        )

    if (
        not args.from_file
        and not args.k8s_fetchai_defaults
        and (
            args.k8s_public_hostname is None
            or args.k8s_namespace is None
            or args.k8s_template_files_dir is None
        )
    ):
        parser.error(
            "--k8s-public-hostname, --k8s-namespace, --k8s-template-files-dir are required when --k8s-fetchai-defaults is not set"
        )

    if (
        not args.from_file
        and not args.docker_fetchai_defaults
        and not args.docker_fetchai_defaults_dev
        and (
            args.docker_file is None
            or args.docker_ctx is None
            or args.docker_image is None
            or args.docker_registry is None
        )
    ):
        parser.error(
            "--docker-file, --docker-ctx, --docker-image, --docker-registry are required when --docker-fetchai-defaults[-dev] is not set"
        )

    if args.docker_fetchai_defaults and args.docker_fetchai_defaults_dev:
        parser.error(
            "--docker-fetchai-defaults and --docker-fetchai-defaults-dev are mutually exclusive"
        )

    if args.generate_only and (args.delete_deployment or args.from_file):
        parser.error("--generate-only can not be used with --delete or --from-file")

    return args


def main():
    """K8s deploy acn node"""

    args = parse_commandline()

    pod_deployment: Optional[K8sPodDeployment] = None

    if args.from_file:
        pod_deployment = K8sPodDeployment([Path(args.from_file)], None)
    else:
        dargs: List[Any] = [
            args.key,
            args.port,
            args.delegate_port,
            args.monitoring_port,
        ]

        dargs.append(
            args.entry_peers_maddrs if args.entry_peers_maddrs is not None else ""
        )

        docker_config: List[str] = []
        if args.docker_fetchai_defaults_dev:
            docker_config = [
                DOCKER_FETCHAI_DEFAULT_FILE_DEV,
                DOCKER_FETCHAI_DEFAULT_CTX_DEV,
                DOCKER_FETCHAI_DEFAULT_IMG,
                DOCKER_FETCHAI_DEFAULT_REGISTRY,
            ]
        elif args.docker_fetchai_defaults:
            docker_config = [
                DOCKER_FETCHAI_DEFAULT_FILE,
                DOCKER_FETCHAI_DEFAULT_CTX,
                DOCKER_FETCHAI_DEFAULT_IMG,
                DOCKER_FETCHAI_DEFAULT_REGISTRY,
            ]
        if args.docker_file is not None:
            docker_config[0] = args.docker_file
        if args.docker_ctx is not None:
            docker_config[1] = args.docker_ctx
        if args.docker_image is not None:
            docker_config[2] = args.docker_image
        if args.docker_registry is not None:
            docker_config[3] = args.docker_registry

        dargs.extend(docker_config)

        k8s_config: List[str] = []
        if args.k8s_fetchai_defaults:
            k8s_config = [
                K8S_FETCHAI_DEFAULT_PUBLIC_HOST,
                K8S_FETCHAI_DEFAULT_NAMESPACE,
                K8S_FETCHAI_DEFAULT_PUBLIC_TEMPLATE_DIR,
            ]
        if args.k8s_public_hostname is not None:
            k8s_config[0] = args.k8s_public_hostname
        if args.k8s_namespace is not None:
            k8s_config[1] = args.k8s_namespace
        if args.k8s_template_files_dir is not None:
            k8s_config[2] = args.k8s_template_files_dir

        dargs.extend(k8s_config)

        pod_deployment = AcnK8sPodConfig(
            dargs[0],
            dargs[1],
            dargs[2],
            dargs[3],
            dargs[4],
            dargs[5],
            dargs[6],
            dargs[7],
            dargs[8],
            dargs[9],
            dargs[10],
            dargs[11],
        ).generate_deployment()

    if args.generate_only:
        return

    try:
        if args.delete_deployment:
            pod_deployment.delete()
        else:
            pod_deployment.deploy()
    except Exception as e:
        raise e


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This package contains a handler to process HTTP responses."""

from typing import cast

from aea.protocols.base import Message
from aea.skills.base import Handler

from packages.fetchai.protocols.http.message import HttpMessage
from packages.fetchai.skills.simple_data_request.dialogues import (
    HttpDialogue,
    HttpDialogues,
)


class HttpHandler(Handler):
    """This class represents the handler for HTTP messages."""

    SUPPORTED_PROTOCOL = HttpMessage.protocol_id

    def __init__(self, **kwargs):
        """Initialize the handler."""
        self.shared_state_key = kwargs.pop("shared_state_key", None)
        if self.shared_state_key is None:
            raise ValueError("No shared_state_key provided!")
        super().__init__(**kwargs)

    def setup(self) -> None:
        """
        Implement the setup.

        :return: None
        """
        pass

    def handle(self, message: Message) -> None:
        """
        Implement the reaction to an envelope.

        :param message: the message
        :return: None
        """
        http_msg = cast(HttpMessage, message)

        # recover dialogue
        http_dialogues = cast(HttpDialogues, self.context.http_dialogues)
        http_dialogue = cast(HttpDialogue, http_dialogues.update(http_msg))
        if http_dialogue is None:
            self._handle_unidentified_dialogue(http_msg)
            return

        # handle message (only process responses, we do not expect requests)
        if http_msg.performative == HttpMessage.Performative.RESPONSE:
            self._handle_response(http_msg, http_dialogue)
        else:
            self._handle_invalid(http_msg, http_dialogue)

    def _handle_unidentified_dialogue(self, http_msg: HttpMessage) -> None:
        """
        Handle an unidentified dialogue.

        :param http_msg: the message
        :return: None
        """
        self.context.logger.info(
            "received invalid http message={}, unidentified dialogue.".format(http_msg)
        )

    def _handle_response(
        self, http_msg: HttpMessage, http_dialogue: HttpDialogue
    ) -> None:
        """
        Handle a http response.

        :param http_msg: the message
        :param http_dialogue: the dialogue object
        :return: None
        """
        self.context.logger.debug(
            "received http response={} in dialogue={}.".format(http_msg, http_dialogue)
        )
        data_received = http_msg.body

        # save the data in the shared state to make it accessible to other skills
        self.context.logger.info(
            "updating shared_state with received data={!r}!".format(data_received)
        )
        self.context.shared_state[self.shared_state_key] = data_received

    def _handle_invalid(
        self, http_msg: HttpMessage, http_dialogue: HttpDialogue
    ) -> None:
        """
        Handle a http message of invalid performative.

        :param http_msg: the message
        :param http_dialogue: the dialogue object
        :return: None
        """
        self.context.logger.warning(
            "cannot handle http message of performative={} in dialogue={}.".format(
                http_msg.performative, http_dialogue
            )
        )

    def teardown(self) -> None:
        """
        Implement the handler teardown.

        :return: None
        """
        pass
